import numpy as np
import torch

from agilerl.modules.gpt import (
    MLP,
    CausalSelfAttention,
    EvolvableGPT,
    PositionalEncoding,
    TokenEmbedding,
)

#### TESTING EvolvableGPT CLASS ####


# The model can be initialized with default parameters.
def test_default_parameters_initialization():
    model = EvolvableGPT()
    assert model.n_layer == 12
    assert model.vocab_size == 50257
    assert model.n_embd == 768
    assert model.n_head == 12
    assert model.dim_feedfwd == 3072
    assert model.block_size == 1024
    assert model.dropout == 0.0
    assert model.activation == "GELU"
    assert model.layer_norm_eps == 1e-05
    assert model.min_layers == 8
    assert model.max_layers == 16
    assert model.bias is True
    assert model.device == "cpu"


# The model can be initialized with custom parameters.
def test_custom_parameters_initialization():
    model = EvolvableGPT(
        n_layer=6,
        vocab_size=10000,
        n_embd=512,
        n_head=8,
        dim_feedfwd=2048,
        block_size=512,
        dropout=0.2,
        activation="ReLU",
        layer_norm_eps=1e-06,
        min_layers=4,
        max_layers=10,
        bias=False,
        device="cpu",
    )
    assert model.n_layer == 6
    assert model.vocab_size == 10000
    assert model.n_embd == 512
    assert model.n_head == 8
    assert model.dim_feedfwd == 2048
    assert model.block_size == 512
    assert model.dropout == 0.2
    assert model.activation == "ReLU"
    assert model.layer_norm_eps == 1e-06
    assert model.min_layers == 4
    assert model.max_layers == 10
    assert model.bias is False
    assert model.device == "cpu"


# The model can be loaded from a pretrained GPT model.
def test_pretrained_model_loading():
    model = EvolvableGPT.from_pretrained("gpt2", override_args={"dropout": 0.1})
    assert model.n_layer == 12
    assert model.vocab_size == 50257
    assert model.n_embd == 768
    assert model.n_head == 12
    assert model.dim_feedfwd == 3072
    assert model.block_size == 1024
    assert model.dropout == 0.1
    assert model.activation == "GELU"
    assert model.layer_norm_eps == 1e-05
    assert model.min_layers == 8
    assert model.max_layers == 16
    assert model.bias is True
    assert model.device == "cpu"


# Configures optimizers for EvoGPT
def test_configure_optimizers():
    weight_decay = 0.0
    learning_rate = 1e-6
    betas = (0.9, 0.999)

    model = EvolvableGPT()
    optimizer = model.configure_optimizers(weight_decay, learning_rate, betas, "cpu")

    assert isinstance(optimizer, torch.optim.AdamW)


# The model can handle a sequence of tokens with a target for training.
def test_sequence_with_target_handling():
    model = EvolvableGPT()
    input_sequence = torch.randint(0, model.vocab_size, (1, model.block_size))
    target_sequence = torch.randint(0, model.vocab_size, (1, model.block_size))
    logits, all_hidden_states, presents, loss = model(
        input_sequence, targets=target_sequence
    )
    assert logits.shape[1] == model.block_size
    assert loss is not None

    tok_emb = model.transformer.wte(input_sequence)
    t = tok_emb.size(-2)
    # past_key_values = tuple([None] * model.n_layer)
    past_length = 0
    pos = torch.arange(
        past_length, t + past_length, dtype=torch.long, device=tok_emb.device
    ).unsqueeze(0)
    logits, all_hidden_states, presents, loss = model(tok_emb=tok_emb, pos=pos)
    assert logits.shape[1] == model.block_size
    assert loss is None


# The model can handle a sequence of tokens without a target for generation.
def test_sequence_without_target_handling():
    model = EvolvableGPT()
    input_sequence = torch.randint(0, model.vocab_size, (1, model.block_size))
    generated_sequence = model.generate(input_sequence, max_new_tokens=10, top_k=3)
    assert generated_sequence.shape[1] == model.block_size + 10


# Decrease block size successfully
def test_decrease_block_size_successfully():
    block_size = 512
    model = EvolvableGPT()
    assert model.block_size == 1024

    # Manually override flash attention
    for block in model.transformer.h:
        block.attn.flash = False
        block.attn.register_buffer(
            "attention_bias",
            torch.tril(torch.ones(block_size, block_size)).view(
                1, 1, block_size, block_size
            ),
        )

    model.crop_block_size(block_size)
    assert model.block_size == block_size
    assert model.transformer.wpe.weight.shape[0] == block_size
    for block in model.transformer.h:
        if hasattr(block.attn, "attention_bias"):
            assert block.attn.attention_bias.shape[2] == block_size
            assert block.attn.attention_bias.shape[3] == block_size


# The model can estimate the MFU (Million Floating Point Operations per Second).
def test_estimate_mfu():
    model = EvolvableGPT()
    fwdbwd_per_iter = 10
    dt = 0.1
    mfu = model.estimate_mfu(fwdbwd_per_iter, dt)
    assert isinstance(mfu, float)
    assert mfu >= 0.0


# The model can generate new tokens based on an input sequence.
def test_generate_new_tokens():
    model = EvolvableGPT()
    idx = torch.tensor([[0, 1, 2, 3, 4]])
    max_new_tokens = 5
    temperature = 1.0
    top_k = None

    generated_tokens = model.generate(idx, max_new_tokens, temperature, top_k)

    assert generated_tokens.size() == (1, 10)


# The model can forward pass a sequence of tokens and return the logits.
def test_forward_pass():
    model = EvolvableGPT()
    input_tokens = torch.tensor([[1, 2, 3, 4, 5]])
    B = 1  # Batch size
    C = 768  # Embedding dim

    k = torch.randn((B, 12, 0, C // 12))
    v = torch.randn((B, 12, 0, C // 12))
    past_kv = [(k, v) for _ in range(12)]
    logits, _, _, _ = model(input_tokens, past_key_values=past_kv)
    assert logits.shape == (1, 5, model.vocab_size)


# The model can count the number of parameters.
def test_count_parameters():
    model = EvolvableGPT()
    num_params = model.get_num_params()
    assert isinstance(num_params, int)
    assert num_params >= 0


# Adds a layer to transformer
def test_add_layer():
    model = EvolvableGPT()
    initial_n_layer = model.n_layer
    model.add_layer()
    assert model.n_layer == initial_n_layer + 1
    assert len(model.transformer.h) == initial_n_layer + 1


# Removes a layer to transformer
def test_remove_layer():
    model = EvolvableGPT()
    initial_n_layer = model.n_layer
    model.remove_layer()
    assert model.n_layer == initial_n_layer - 1
    assert len(model.transformer.h) == initial_n_layer - 1


# Adds nodes to transformer
def test_add_nodes():
    model = EvolvableGPT()
    initial_dim_feedfwd = model.dim_feedfwd
    model.add_node()
    assert model.dim_feedfwd > initial_dim_feedfwd
    for block in model.transformer.h:
        assert block.mlp.hidden_size[0] > initial_dim_feedfwd


# Removes nodes to transformer
def test_remove_nodes():
    model = EvolvableGPT()
    initial_dim_feedfwd = model.dim_feedfwd
    model.remove_node()
    assert model.dim_feedfwd < initial_dim_feedfwd
    for block in model.transformer.h:
        assert block.mlp.hidden_size[0] < initial_dim_feedfwd


# The model can clone itself.
def test_model_clone():
    model = EvolvableGPT()
    clone = model.clone()
    assert isinstance(clone, EvolvableGPT)
    assert clone.n_layer == model.n_layer
    assert clone.vocab_size == model.vocab_size
    assert clone.n_embd == model.n_embd
    assert clone.n_head == model.n_head
    assert clone.dim_feedfwd == model.dim_feedfwd
    assert clone.block_size == model.block_size
    assert clone.dropout == model.dropout
    assert clone.activation == model.activation
    assert clone.layer_norm_eps == model.layer_norm_eps
    assert clone.min_layers == model.min_layers
    assert clone.max_layers == model.max_layers
    assert clone.bias == model.bias
    assert clone.device == model.device


#### TESTING CAUSAL SELF ATTENTION CLASS ####
def test_causal_self_attention_forward():
    block_size = 1024
    attn = CausalSelfAttention(768, 12, True, 0.1, block_size)
    attn.flash = False
    attn.register_buffer(
        "attention_bias",
        torch.tril(torch.ones(block_size, block_size)).view(
            1, 1, block_size, block_size
        ),
    )
    B = 4  # Batch size
    T = 128  # Sequence length
    C = 768  # Embedding dim
    x = torch.randint(0, 32, (B, T, C)).float()

    k = torch.randn((B, 12, 0, C // 12))
    v = torch.randn((B, 12, 0, C // 12))
    layer_past = (k, v)

    y, present = attn(x, layer_past=layer_past)

    assert isinstance(y, torch.Tensor)
    assert isinstance(present[0], torch.Tensor)
    assert isinstance(present[1], torch.Tensor)


#### TESTING MLP CLASS ####
def test_forward_mlp():
    input_array = np.random.rand(1, 32)
    model = MLP(32, 0.1, 64)

    output = model(input_array)

    assert isinstance(output, torch.Tensor)
    assert output.shape == (1, 32)


#### TESTING POSITIONAL ENCODING CLASS ####
def test_pos_encoding():
    input_tensor = torch.LongTensor([[1, 2, 4, 5]])
    pos = PositionalEncoding(10, 3)
    enc = pos(input_tensor)

    assert enc.shape == (1, 4, 3)


#### TESTING TOKEN EMBEDDING CLASS ####
def test_tok_embedding():
    input_tensor = torch.LongTensor([[1, 2, 4, 5]])
    tok = TokenEmbedding(10, 3)
    emb = tok(input_tensor)

    assert emb.shape == (1, 4, 3)



import copy

import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces
from tensordict import TensorDict

from agilerl.algorithms.neural_ucb_bandit import NeuralUCB
from agilerl.modules import EvolvableCNN, EvolvableMLP, EvolvableMultiInput
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import assert_state_dicts_equal


class DummyNeuralUCB(NeuralUCB):
    def __init__(
        self,
        observation_space,
        action_space,
        net_config=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            observation_space=observation_space,
            action_space=action_space,
            net_config=net_config,
            *args,
            **kwargs,
        )

        self.tensor_test = torch.randn(1)
        self.numpy_test = np.random.rand(1)


class DummyBanditEnv:
    def __init__(self, state_size, arms):
        self.arms = arms
        self.state_size = (arms,) + state_size
        self.n_envs = 1

    def reset(self):
        return np.random.rand(*self.state_size)

    def step(self, action):
        return (
            np.random.rand(*self.state_size),
            np.random.rand(self.n_envs),
        )


@pytest.mark.parametrize(
    "observation_space, encoder_cls",
    [
        ("vector_space", EvolvableMLP),
        ("image_space", EvolvableCNN),
        ("dict_space", EvolvableMultiInput),
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_initialize_bandit(observation_space, encoder_cls, accelerator, request):
    action_space = spaces.Discrete(2)
    observation_space = request.getfixturevalue(observation_space)
    device = accelerator.device if accelerator else "cpu"
    bandit = NeuralUCB(observation_space, action_space, accelerator=accelerator)

    assert bandit.observation_space == observation_space
    assert bandit.action_space == action_space
    assert bandit.batch_size == 64
    assert bandit.lr == 0.001
    assert bandit.learn_step == 2
    assert bandit.gamma == 1.0
    assert bandit.lamb == 1.0
    assert bandit.reg == 0.000625
    assert bandit.mut is None
    assert bandit.device == device
    assert bandit.accelerator == accelerator
    assert bandit.index == 0
    assert bandit.scores == []
    assert bandit.fitness == []
    assert bandit.steps == [0]
    assert isinstance(bandit.actor.encoder, encoder_cls)
    expected_optimizer = AcceleratedOptimizer if accelerator else optim.Adam
    assert isinstance(bandit.optimizer.optimizer, expected_optimizer)
    assert isinstance(bandit.criterion, nn.MSELoss)
    bandit.clean_up()


# Can initialize NeuralUCB with an actor network
# TODO: Will be deprecated in the future
@pytest.mark.parametrize(
    "observation_space, actor_network, input_tensor",
    [
        ("vector_space", "simple_mlp", torch.randn(1, 4)),
        ("image_space", "simple_cnn", torch.randn(1, 3, 32, 32)),
    ],
)
def test_initialize_bandit_with_actor_network(
    observation_space, discrete_space, actor_network, input_tensor, request
):
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)

    bandit = NeuralUCB(observation_space, discrete_space, actor_network=actor_network)

    assert bandit.observation_space == observation_space
    assert bandit.action_space == discrete_space
    assert bandit.batch_size == 64
    assert bandit.lr == 0.001
    assert bandit.learn_step == 2
    assert bandit.gamma == 1.0
    assert bandit.lamb == 1.0
    assert bandit.reg == 0.000625
    assert bandit.mut is None
    assert bandit.device == "cpu"
    assert bandit.accelerator is None
    assert bandit.index == 0
    assert bandit.scores == []
    assert bandit.fitness == []
    assert bandit.steps == [0]
    assert isinstance(bandit.optimizer.optimizer, optim.Adam)
    assert isinstance(bandit.criterion, nn.MSELoss)
    bandit.clean_up()


def test_initialize_bandit_with_incorrect_actor_network(vector_space, discrete_space):
    actor_network = nn.Sequential(nn.Linear(vector_space.shape[0], discrete_space.n))

    with pytest.raises(TypeError) as e:
        bandit = NeuralUCB(vector_space, discrete_space, actor_network=actor_network)

        assert bandit
        assert (
            e
            == "'actor_network' argument is of type {type(actor_network)}, but must be of type EvolvableModule."
        )


def test_initialize_bandit_with_evo_nets(vector_space, discrete_space):
    actor_network = EvolvableMLP(
        num_inputs=vector_space.shape[0],
        num_outputs=1,
        hidden_size=[64, 64],
        layer_norm=False,
    )

    bandit = NeuralUCB(vector_space, discrete_space, actor_network=actor_network)
    assert bandit.batch_size == 64
    assert bandit.lr == 0.001
    assert bandit.learn_step == 2
    assert bandit.gamma == 1.0
    assert bandit.lamb == 1.0
    assert bandit.reg == 0.000625
    assert bandit.mut is None
    assert bandit.device == "cpu"
    assert bandit.accelerator is None
    assert bandit.index == 0
    assert bandit.scores == []
    assert bandit.fitness == []
    assert bandit.steps == [0]
    assert isinstance(bandit.optimizer.optimizer, optim.Adam)
    assert isinstance(bandit.criterion, nn.MSELoss)
    bandit.clean_up()


def test_initialize_neuralucb_with_incorrect_actor_net_type(
    vector_space, discrete_space
):
    actor_network = "dummy"
    with pytest.raises(TypeError) as a:
        bandit = NeuralUCB(vector_space, discrete_space, actor_network=actor_network)
        assert bandit
        assert (
            str(a.value)
            == f"'actor_network' argument is of type {type(actor_network)}, but must be of type EvolvableMLP, EvolvableCNN or MakeEvolvable"
        )


# Returns the expected action when given a state observation and epsilon=0 or 1.
def test_returns_expected_action(vector_space, discrete_space):
    bandit = NeuralUCB(vector_space, discrete_space)
    state = np.array([1, 2, 3, 4])

    action_mask = None

    action = bandit.get_action(state, action_mask)

    assert action.is_integer()
    assert action >= 0 and action < discrete_space.n
    bandit.clean_up()


# Returns the expected action when given a state observation and action mask.
def test_returns_expected_action_mask(vector_space, discrete_space):
    accelerator = Accelerator()

    bandit = NeuralUCB(vector_space, discrete_space, accelerator=accelerator)
    state = np.array([1, 2, 3, 4])

    action_mask = np.array([0, 1])

    action = bandit.get_action(state, action_mask)

    assert action.is_integer()
    assert action == 1
    bandit.clean_up()


# learns from experiences and updates network parameters
@pytest.mark.parametrize("observation_space", ["vector_space", "image_space"])
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_learns_from_experiences(
    observation_space, discrete_space, accelerator, request
):
    observation_space = request.getfixturevalue(observation_space)
    batch_size = 64

    # Create an instance of the NeuralTS class
    bandit = NeuralUCB(
        observation_space,
        discrete_space,
        batch_size=batch_size,
        accelerator=accelerator,
    )

    # Create a batch of experiences
    states = torch.randn(batch_size, *observation_space.shape)
    rewards = torch.randn((batch_size, 1))

    experiences = TensorDict(
        {"obs": states, "reward": rewards},
        batch_size=[batch_size],
        device=bandit.device,
    )

    # Copy state dict before learning - should be different to after updating weights
    actor = bandit.actor
    actor_pre_learn_sd = str(copy.deepcopy(bandit.actor.state_dict()))

    # Call the learn method
    loss = bandit.learn(experiences)

    assert isinstance(loss, float)
    assert loss >= 0.0
    assert actor == bandit.actor
    assert actor_pre_learn_sd != str(bandit.actor.state_dict())
    bandit.clean_up()


# Runs algorithm test loop
@pytest.mark.parametrize("observation_space", ["vector_space", "image_space"])
def test_algorithm_test_loop(observation_space, discrete_space, request):
    observation_space = request.getfixturevalue(observation_space)

    env = DummyBanditEnv(state_size=observation_space.shape, arms=discrete_space.n)

    agent = NeuralUCB(observation_space=observation_space, action_space=discrete_space)
    mean_score = agent.test(env, max_steps=10)
    assert isinstance(mean_score, float)
    agent.clean_up()


# Clones the agent and returns an identical agent.
@pytest.mark.parametrize("observation_space", ["vector_space"])
def test_clone_returns_identical_agent(observation_space, discrete_space, request):
    observation_space = request.getfixturevalue(observation_space)
    bandit = DummyNeuralUCB(observation_space, discrete_space)
    bandit.tensor_attribute = torch.randn(1)
    bandit.numpy_attribute = np.random.rand(1)
    clone_agent = bandit.clone()

    assert clone_agent.observation_space == bandit.observation_space
    assert clone_agent.action_space == bandit.action_space
    assert clone_agent.batch_size == bandit.batch_size
    assert clone_agent.lr == bandit.lr
    assert clone_agent.learn_step == bandit.learn_step
    assert clone_agent.gamma == bandit.gamma
    assert clone_agent.mut == bandit.mut
    assert clone_agent.device == bandit.device
    assert clone_agent.accelerator == bandit.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), bandit.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), bandit.optimizer.state_dict()
    )
    assert clone_agent.fitness == bandit.fitness
    assert clone_agent.steps == bandit.steps
    assert clone_agent.scores == bandit.scores
    assert clone_agent.tensor_attribute == bandit.tensor_attribute
    assert clone_agent.tensor_test == bandit.tensor_test
    assert clone_agent.numpy_attribute == bandit.numpy_attribute
    assert clone_agent.numpy_test == bandit.numpy_test
    bandit.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    bandit = NeuralUCB(observation_space, discrete_space, accelerator=accelerator)
    clone_agent = bandit.clone()

    assert clone_agent.observation_space == bandit.observation_space
    assert clone_agent.action_space == bandit.action_space
    assert clone_agent.batch_size == bandit.batch_size
    assert clone_agent.lr == bandit.lr
    assert clone_agent.learn_step == bandit.learn_step
    assert clone_agent.gamma == bandit.gamma
    assert clone_agent.mut == bandit.mut
    assert clone_agent.device == bandit.device
    assert clone_agent.accelerator == bandit.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), bandit.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), bandit.optimizer.state_dict()
    )
    assert clone_agent.fitness == bandit.fitness
    assert clone_agent.steps == bandit.steps
    assert clone_agent.scores == bandit.scores
    bandit.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    bandit = NeuralUCB(
        observation_space, discrete_space, accelerator=accelerator, wrap=False
    )
    clone_agent = bandit.clone(wrap=False)

    assert clone_agent.observation_space == bandit.observation_space
    assert clone_agent.action_space == bandit.action_space
    assert clone_agent.batch_size == bandit.batch_size
    assert clone_agent.lr == bandit.lr
    assert clone_agent.learn_step == bandit.learn_step
    assert clone_agent.gamma == bandit.gamma
    assert clone_agent.mut == bandit.mut
    assert clone_agent.device == bandit.device
    assert clone_agent.accelerator == bandit.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), bandit.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), bandit.optimizer.state_dict()
    )
    assert clone_agent.fitness == bandit.fitness
    assert clone_agent.steps == bandit.steps
    assert clone_agent.scores == bandit.scores
    bandit.clean_up()
    clone_agent.clean_up()


def test_clone_new_index(vector_space, discrete_space):
    bandit = NeuralUCB(vector_space, discrete_space)
    clone_agent = bandit.clone(index=100)
    assert clone_agent.index == 100
    bandit.clean_up()
    clone_agent.clean_up()


def test_clone_after_learning(vector_space, discrete_space):
    batch_size = 4
    states = torch.randn(batch_size, vector_space.shape[0])
    rewards = torch.rand(batch_size, 1)
    experiences = TensorDict(
        {"obs": states, "reward": rewards},
        batch_size=[batch_size],
    )
    bandit = NeuralUCB(vector_space, discrete_space, batch_size=batch_size)
    bandit.learn(experiences)
    clone_agent = bandit.clone()

    assert clone_agent.observation_space == bandit.observation_space
    assert clone_agent.action_space == discrete_space
    assert clone_agent.batch_size == bandit.batch_size
    assert clone_agent.lr == bandit.lr
    assert clone_agent.learn_step == bandit.learn_step
    assert clone_agent.gamma == bandit.gamma
    assert clone_agent.mut == bandit.mut
    assert clone_agent.device == bandit.device
    assert clone_agent.accelerator == bandit.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), bandit.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), bandit.optimizer.state_dict()
    )
    assert clone_agent.fitness == bandit.fitness
    assert clone_agent.steps == bandit.steps
    assert clone_agent.scores == bandit.scores
    bandit.clean_up()
    clone_agent.clean_up()


# TODO: Will be deprecated in the future
@pytest.mark.parametrize(
    "observation_space, actor_network, input_tensor",
    [
        ("vector_space", "simple_mlp", torch.randn(1, 4)),
    ],
)
def test_clone_with_make_evo(
    observation_space, discrete_space, actor_network, input_tensor, request
):
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)

    bandit = NeuralUCB(observation_space, discrete_space, actor_network=actor_network)
    clone_agent = bandit.clone()

    assert clone_agent.observation_space == bandit.observation_space
    assert clone_agent.action_space == discrete_space
    assert clone_agent.batch_size == bandit.batch_size
    assert clone_agent.lr == bandit.lr
    assert clone_agent.learn_step == bandit.learn_step
    assert clone_agent.gamma == bandit.gamma
    assert clone_agent.mut == bandit.mut
    assert clone_agent.device == bandit.device
    assert clone_agent.accelerator == bandit.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), bandit.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), bandit.optimizer.state_dict()
    )
    assert clone_agent.fitness == bandit.fitness
    assert clone_agent.steps == bandit.steps
    assert clone_agent.scores == bandit.scores
    bandit.clean_up()
    clone_agent.clean_up()



import copy

import numpy as np
import pytest
import torch
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from tensordict import TensorDict

from agilerl.algorithms.dqn_rainbow import RainbowDQN
from agilerl.modules import EvolvableCNN, EvolvableMLP, EvolvableMultiInput
from agilerl.networks.q_networks import RainbowQNetwork
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import (
    assert_not_equal_state_dict,
    assert_state_dicts_equal,
    get_experiences_batch,
    get_sample_from_space,
)


class DummyRainbowDQN(RainbowDQN):
    def __init__(self, observation_space, action_space, *args, **kwargs):
        super().__init__(observation_space, action_space, *args, **kwargs)

        self.tensor_test = torch.randn(1)


class DummyEnv:
    def __init__(self, state_size, vect=True, num_envs=2):
        self.state_size = state_size
        self.vect = vect
        if self.vect:
            self.state_size = (num_envs,) + self.state_size
            self.n_envs = num_envs
            self.num_envs = num_envs
        else:
            self.n_envs = 1

    def reset(self):
        return np.random.rand(*self.state_size), {}

    def step(self, action):
        return (
            np.random.rand(*self.state_size),
            np.random.randint(0, 5, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            {},
        )


# initialize DQN with valid parameters
@pytest.mark.parametrize(
    "observation_space, encoder_cls",
    [
        ("vector_space", EvolvableMLP),
        ("image_space", EvolvableCNN),
        ("dict_space", EvolvableMultiInput),
        ("multidiscrete_space", EvolvableMLP),
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_initialize_dqn(
    observation_space, discrete_space, encoder_cls, accelerator, request
):
    observation_space = request.getfixturevalue(observation_space)
    dqn = RainbowDQN(observation_space, discrete_space, accelerator=accelerator)

    expected_device = accelerator.device if accelerator else "cpu"
    assert dqn.observation_space == observation_space
    assert dqn.action_space == discrete_space
    assert dqn.batch_size == 64
    assert dqn.lr == 0.0001
    assert dqn.learn_step == 5
    assert dqn.gamma == 0.99
    assert dqn.tau == 0.001
    assert dqn.mut is None
    assert dqn.device == expected_device
    assert dqn.accelerator == accelerator
    assert dqn.index == 0
    assert dqn.scores == []
    assert dqn.fitness == []
    assert dqn.steps == [0]
    # assert dqn.actor_network is None
    assert isinstance(dqn.actor.encoder, encoder_cls)
    assert isinstance(dqn.actor_target.encoder, encoder_cls)
    expected_opt_cls = AcceleratedOptimizer if accelerator else optim.Adam
    assert isinstance(dqn.optimizer.optimizer, expected_opt_cls)
    dqn.clean_up()


@pytest.mark.parametrize(
    "observation_space, encoder_cls",
    [
        ("vector_space", EvolvableMLP),
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_initialize_dqn_with_actor_network_evo_net(
    observation_space, discrete_space, encoder_cls, accelerator, request
):
    observation_space = request.getfixturevalue(observation_space)
    support = torch.linspace(0, 1, 51)
    device = accelerator.device if accelerator else "cpu"
    actor_network = RainbowQNetwork(
        observation_space=observation_space,
        action_space=discrete_space,
        support=support,
        device=device,
    )

    # Create an instance of the RainbowDQN class
    dqn = RainbowDQN(
        observation_space,
        discrete_space,
        actor_network=actor_network,
        accelerator=accelerator,
    )

    assert dqn.observation_space == observation_space
    assert dqn.action_space == discrete_space
    assert dqn.batch_size == 64
    assert dqn.lr == 0.0001
    assert dqn.learn_step == 5
    assert dqn.gamma == 0.99
    assert dqn.tau == 0.001
    assert dqn.mut is None
    assert dqn.device == device
    assert dqn.accelerator == accelerator
    assert dqn.index == 0
    assert dqn.scores == []
    assert dqn.fitness == []
    assert dqn.steps == [0]

    assert isinstance(dqn.actor.encoder, encoder_cls)
    assert isinstance(dqn.actor_target.encoder, encoder_cls)
    if accelerator is not None:
        assert isinstance(dqn.optimizer.optimizer, AcceleratedOptimizer)
    else:
        assert isinstance(dqn.optimizer.optimizer, optim.Adam)
    dqn.clean_up()


def test_initialize_dqn_with_incorrect_actor_net_type(
    vector_space, discrete_space, request
):
    actor_network = "dummy"

    with pytest.raises(TypeError) as a:
        dqn = RainbowDQN(vector_space, discrete_space, actor_network=actor_network)
        assert dqn
        assert (
            str(a.value)
            == f"'actor_network' argument is of type {type(actor_network)}, but must be of type nn.Module."
        )


# Can initialize DQN with an actor network
# TODO: This will be deprecated in the future
@pytest.mark.parametrize(
    "observation_space, actor_network, input_tensor",
    [
        ("vector_space", "simple_mlp", torch.randn(1, 4)),
        (
            "image_space",
            "simple_cnn",
            torch.randn(1, 3, 32, 32),
        ),
    ],
)
def test_initialize_dqn_with_make_evolvable(
    observation_space, discrete_space, actor_network, input_tensor, request
):
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)

    dqn = RainbowDQN(observation_space, discrete_space, actor_network=actor_network)

    assert dqn.observation_space == observation_space
    assert dqn.action_space == discrete_space
    assert dqn.batch_size == 64
    assert dqn.lr == 0.0001
    assert dqn.learn_step == 5
    assert dqn.gamma == 0.99
    assert dqn.tau == 0.001
    assert dqn.mut is None
    assert dqn.device == "cpu"
    assert dqn.accelerator is None
    assert dqn.index == 0
    assert dqn.scores == []
    assert dqn.fitness == []
    assert dqn.steps == [0]
    # assert dqn.actor_network == actor_network
    assert isinstance(dqn.optimizer.optimizer, optim.Adam)


@pytest.mark.parametrize(
    "observation_space",
    [
        "vector_space",
        "discrete_space",
        "image_space",
        "multidiscrete_space",
        "dict_space",
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_returns_expected_action(
    accelerator, observation_space, discrete_space, request
):
    observation_space = request.getfixturevalue(observation_space)
    dqn = RainbowDQN(observation_space, discrete_space, accelerator=accelerator)
    state = get_sample_from_space(observation_space)

    action_mask = None

    action = dqn.get_action(state, action_mask)[0]

    assert action.is_integer()
    assert action >= 0 and action < discrete_space.n

    action_mask = np.array([0, 1])

    action = dqn.get_action(state, action_mask)[0]

    assert action.is_integer()
    assert action == 1
    dqn.clean_up()


def test_returns_expected_action_mask_vectorized(vector_space, discrete_space):
    accelerator = Accelerator()

    dqn = RainbowDQN(vector_space, discrete_space, accelerator=accelerator)
    state = np.array([[1, 2, 4, 5], [2, 3, 5, 1]])

    action_mask = np.array([[0, 1], [1, 0]])

    action = dqn.get_action(state, action_mask)

    assert np.array_equal(action, [1, 0])
    dqn.clean_up()


@pytest.mark.parametrize(
    "observation_space",
    [
        "discrete_space",
        "vector_space",
        "dict_space",
        "multidiscrete_space",
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
# learns from experiences and updates network parameters
def test_learns_from_experiences(
    accelerator, observation_space, discrete_space, request
):
    observation_space = request.getfixturevalue(observation_space)
    torch.autograd.set_detect_anomaly(True)
    batch_size = 64

    # Create an instance of the DQN class
    dqn = RainbowDQN(
        observation_space,
        discrete_space,
        batch_size=batch_size,
        accelerator=accelerator,
    )

    # Create a batch of experiences
    device = accelerator.device if accelerator else "cpu"
    experiences = get_experiences_batch(
        observation_space, discrete_space, batch_size, device
    )

    # Copy state dict before learning - should be different to after updating weights
    actor = dqn.actor
    actor_target = dqn.actor_target
    actor_pre_learn_sd = copy.deepcopy(dqn.actor.state_dict())
    actor_target_pre_learn_sd = copy.deepcopy(dqn.actor_target.state_dict())

    # Call the learn method
    loss, new_idxs, new_priorities = dqn.learn(experiences, per=False)

    assert loss > 0.0
    assert new_idxs is None
    assert new_priorities is None
    assert actor == dqn.actor
    assert actor_target == dqn.actor_target
    assert_not_equal_state_dict(actor_pre_learn_sd, dqn.actor.state_dict())
    assert_not_equal_state_dict(
        actor_target_pre_learn_sd, dqn.actor_target.state_dict()
    )
    dqn.clean_up()


@pytest.mark.parametrize("accelerator", [None, Accelerator()])
@pytest.mark.parametrize("combined", [True, False])
# learns from experiences and updates network parameters
def test_learns_from_experiences_n_step(
    accelerator, combined, vector_space, discrete_space
):
    batch_size = 64

    # Create an instance of the DQN class
    dqn = RainbowDQN(
        vector_space,
        discrete_space,
        batch_size=batch_size,
        accelerator=accelerator,
        combined_reward=combined,
    )

    # Create a batch of experiences
    # Create a batch of experiences
    states = torch.randn(batch_size, vector_space.shape[0])
    actions = torch.randint(0, discrete_space.n, (batch_size, 1))
    rewards = torch.randn((batch_size, 1))
    next_states = torch.randn(batch_size, vector_space.shape[0])
    dones = torch.randint(0, 2, (batch_size, 1))
    idxs = np.arange(batch_size)
    n_states = torch.randn(batch_size, vector_space.shape[0])
    n_actions = torch.randint(0, discrete_space.n, (batch_size, 1))
    n_rewards = torch.randn((batch_size, 1))
    n_next_states = torch.randn(batch_size, vector_space.shape[0])
    n_dones = torch.randint(0, 2, (batch_size, 1))

    experiences = TensorDict(
        {
            "obs": states,
            "action": actions,
            "reward": rewards,
            "next_obs": next_states,
            "done": dones,
            "idxs": idxs,
        },
        batch_size=[batch_size],
        device=accelerator.device if accelerator else "cpu",
    )

    n_experiences = TensorDict(
        {
            "obs": n_states,
            "action": n_actions,
            "reward": n_rewards,
            "next_obs": n_next_states,
            "done": n_dones,
        },
        batch_size=[batch_size],
        device=accelerator.device if accelerator else "cpu",
    )

    # Copy state dict before learning - should be different to after updating weights
    actor = dqn.actor
    actor_target = dqn.actor_target
    actor_pre_learn_sd = copy.deepcopy(dqn.actor.state_dict())
    actor_target_pre_learn_sd = copy.deepcopy(dqn.actor_target.state_dict())

    # Call the learn method
    loss, new_idxs, new_priorities = dqn.learn(experiences, n_experiences, per=False)

    assert loss > 0.0
    assert new_idxs is not None
    assert new_priorities is None
    assert actor == dqn.actor
    assert actor_target == dqn.actor_target
    assert_not_equal_state_dict(actor_pre_learn_sd, dqn.actor.state_dict())
    assert_not_equal_state_dict(
        actor_target_pre_learn_sd, dqn.actor_target.state_dict()
    )
    dqn.clean_up()


# learns from experiences and updates network parameters
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
@pytest.mark.parametrize("combined", [True, False])
def test_learns_from_experiences_per(
    accelerator, combined, vector_space, discrete_space
):
    batch_size = 64

    # Create an instance of the DQN class
    dqn = RainbowDQN(
        vector_space,
        discrete_space,
        batch_size=batch_size,
        accelerator=accelerator,
        combined_reward=combined,
    )

    # Create a batch of experiences
    states = torch.randn(batch_size, vector_space.shape[0])
    actions = torch.randint(0, discrete_space.n, (batch_size, 1))
    rewards = torch.randn((batch_size, 1))
    next_states = torch.randn(batch_size, vector_space.shape[0])
    dones = torch.randint(0, 2, (batch_size, 1))
    weights = torch.rand(batch_size)
    idxs = torch.from_numpy(np.arange(batch_size))

    experiences = TensorDict(
        {
            "obs": states,
            "action": actions,
            "reward": rewards,
            "next_obs": next_states,
            "done": dones,
            "idxs": idxs,
            "weights": weights,
        },
        batch_size=[batch_size],
        device=accelerator.device if accelerator else "cpu",
    )

    # Copy state dict before learning - should be different to after updating weights
    actor = dqn.actor
    actor_target = dqn.actor_target
    actor_pre_learn_sd = copy.deepcopy(dqn.actor.state_dict())
    actor_target_pre_learn_sd = copy.deepcopy(dqn.actor_target.state_dict())

    # Call the learn method
    loss, new_idxs, new_priorities = dqn.learn(experiences, per=True)

    assert loss > 0.0
    assert isinstance(new_idxs, torch.Tensor)
    assert isinstance(new_priorities, np.ndarray)
    assert torch.equal(new_idxs.cpu(), idxs)
    assert actor == dqn.actor
    assert actor_target == dqn.actor_target
    assert_not_equal_state_dict(actor_pre_learn_sd, dqn.actor.state_dict())
    assert_not_equal_state_dict(
        actor_target_pre_learn_sd, dqn.actor_target.state_dict()
    )
    dqn.clean_up()


# learns from experiences and updates network parameters
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
@pytest.mark.parametrize("combined", [True, False])
def test_learns_from_experiences_per_n_step(
    accelerator, combined, vector_space, discrete_space
):
    batch_size = 64

    # Create an instance of the DQN class
    dqn = RainbowDQN(
        vector_space,
        discrete_space,
        batch_size=batch_size,
        accelerator=accelerator,
        combined_reward=combined,
    )

    # Create a batch of experiences
    states = torch.randn(batch_size, vector_space.shape[0])
    actions = torch.randint(0, discrete_space.n, (batch_size, 1))
    rewards = torch.randn((batch_size, 1))
    next_states = torch.randn(batch_size, vector_space.shape[0])
    dones = torch.randint(0, 2, (batch_size, 1))
    weights = torch.rand(batch_size)
    idxs = torch.from_numpy(np.arange(batch_size))
    n_states = torch.randn(batch_size, vector_space.shape[0])
    n_actions = torch.randint(0, discrete_space.n, (batch_size, 1))
    n_rewards = torch.randn((batch_size, 1))
    n_next_states = torch.randn(batch_size, vector_space.shape[0])
    n_dones = torch.randint(0, 2, (batch_size, 1))

    experiences = TensorDict(
        {
            "obs": states,
            "action": actions,
            "reward": rewards,
            "next_obs": next_states,
            "done": dones,
            "idxs": idxs,
            "weights": weights,
        },
        batch_size=[batch_size],
        device=accelerator.device if accelerator else "cpu",
    )

    n_experiences = TensorDict(
        {
            "obs": n_states,
            "action": n_actions,
            "reward": n_rewards,
            "next_obs": n_next_states,
            "done": n_dones,
        },
        batch_size=[batch_size],
        device=accelerator.device if accelerator else "cpu",
    )

    # Copy state dict before learning - should be different to after updating weights
    actor = dqn.actor
    actor_target = dqn.actor_target
    actor_pre_learn_sd = copy.deepcopy(dqn.actor.state_dict())
    actor_target_pre_learn_sd = copy.deepcopy(dqn.actor_target.state_dict())

    # Call the learn method
    loss, new_idxs, new_priorities = dqn.learn(experiences, n_experiences, per=True)

    assert loss > 0.0
    assert isinstance(new_idxs, torch.Tensor)
    assert isinstance(new_priorities, np.ndarray)
    assert torch.equal(new_idxs.cpu(), idxs)
    assert actor == dqn.actor
    assert actor_target == dqn.actor_target
    assert_not_equal_state_dict(actor_pre_learn_sd, dqn.actor.state_dict())
    assert_not_equal_state_dict(
        actor_target_pre_learn_sd, dqn.actor_target.state_dict()
    )
    dqn.clean_up()


# Updates target network parameters with soft update
def test_soft_update(vector_space, discrete_space):
    net_config = {"encoder_config": {"hidden_size": [64, 64]}}
    batch_size = 64
    lr = 1e-4
    learn_step = 5
    gamma = 0.99
    tau = 1e-3
    mut = None
    actor_network = None
    device = "cpu"
    accelerator = None
    wrap = True

    dqn = RainbowDQN(
        vector_space,
        discrete_space,
        net_config=net_config,
        batch_size=batch_size,
        lr=lr,
        learn_step=learn_step,
        gamma=gamma,
        tau=tau,
        mut=mut,
        actor_network=actor_network,
        device=device,
        accelerator=accelerator,
        wrap=wrap,
    )

    dqn.soft_update()

    eval_params = list(dqn.actor.parameters())
    target_params = list(dqn.actor_target.parameters())
    expected_params = [
        dqn.tau * eval_param + (1.0 - dqn.tau) * target_param
        for eval_param, target_param in zip(eval_params, target_params)
    ]

    assert all(
        torch.allclose(expected_param, target_param)
        for expected_param, target_param in zip(expected_params, target_params)
    )
    dqn.clean_up()


# Runs algorithm test loop
@pytest.mark.parametrize("num_envs", [1, 3])
@pytest.mark.parametrize("observation_space", ["vector_space", "image_space"])
def test_algorithm_test_loop(num_envs, observation_space, discrete_space, request):
    observation_space = request.getfixturevalue(observation_space)
    vect = num_envs > 1
    env = DummyEnv(state_size=observation_space.shape, vect=vect, num_envs=num_envs)
    agent = RainbowDQN(observation_space=observation_space, action_space=discrete_space)
    mean_score = agent.test(env, max_steps=10)
    assert isinstance(mean_score, float)
    agent.clean_up()


# Clones the agent and returns an identical agent.
@pytest.mark.parametrize("observation_space", ["vector_space"])
def test_clone_returns_identical_agent(observation_space, discrete_space, request):
    observation_space = request.getfixturevalue(observation_space)
    dqn = DummyRainbowDQN(observation_space, discrete_space)
    dqn.fitness = [200, 200, 200]
    dqn.scores = [94, 94, 94]
    dqn.steps = [2500]
    dqn.tensor_attribute = torch.randn(1)
    clone_agent = dqn.clone()

    assert clone_agent.observation_space == dqn.observation_space
    assert clone_agent.action_space == dqn.action_space
    # assert clone_agent.actor_network == dqn.actor_network
    assert clone_agent.batch_size == dqn.batch_size
    assert clone_agent.lr == dqn.lr
    assert clone_agent.learn_step == dqn.learn_step
    assert clone_agent.gamma == dqn.gamma
    assert clone_agent.tau == dqn.tau
    assert clone_agent.mut == dqn.mut
    assert clone_agent.device == dqn.device
    assert clone_agent.accelerator == dqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), dqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), dqn.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), dqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == dqn.fitness
    assert clone_agent.steps == dqn.steps
    assert clone_agent.scores == dqn.scores
    assert clone_agent.tensor_attribute == dqn.tensor_attribute
    assert clone_agent.tensor_test == dqn.tensor_test
    dqn.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    dqn = RainbowDQN(observation_space, discrete_space, accelerator=accelerator)
    clone_agent = dqn.clone()

    assert clone_agent.observation_space == dqn.observation_space
    assert clone_agent.action_space == dqn.action_space
    assert clone_agent.batch_size == dqn.batch_size
    assert clone_agent.lr == dqn.lr
    assert clone_agent.learn_step == dqn.learn_step
    assert clone_agent.gamma == dqn.gamma
    assert clone_agent.tau == dqn.tau
    assert clone_agent.mut == dqn.mut
    assert clone_agent.device == dqn.device
    assert clone_agent.accelerator == dqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), dqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), dqn.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), dqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == dqn.fitness
    assert clone_agent.steps == dqn.steps
    assert clone_agent.scores == dqn.scores
    dqn.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    dqn = RainbowDQN(
        observation_space, discrete_space, accelerator=accelerator, wrap=False
    )
    clone_agent = dqn.clone(wrap=False)

    assert clone_agent.observation_space == dqn.observation_space
    assert clone_agent.action_space == dqn.action_space
    # assert clone_agent.actor_network == dqn.actor_network
    assert clone_agent.batch_size == dqn.batch_size
    assert clone_agent.lr == dqn.lr
    assert clone_agent.learn_step == dqn.learn_step
    assert clone_agent.gamma == dqn.gamma
    assert clone_agent.tau == dqn.tau
    assert clone_agent.mut == dqn.mut
    assert clone_agent.device == dqn.device
    assert clone_agent.accelerator == dqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), dqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), dqn.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), dqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == dqn.fitness
    assert clone_agent.steps == dqn.steps
    assert clone_agent.scores == dqn.scores
    dqn.clean_up()
    clone_agent.clean_up()


def test_clone_new_index(vector_space, discrete_space):
    dqn = RainbowDQN(vector_space, discrete_space)
    clone_agent = dqn.clone(index=100)

    assert clone_agent.index == 100
    dqn.clean_up()
    clone_agent.clean_up()


def test_clone_after_learning(vector_space, discrete_space):
    batch_size = 8
    rainbow_dqn = RainbowDQN(vector_space, discrete_space, batch_size=batch_size)

    experiences = get_experiences_batch(vector_space, discrete_space, batch_size)
    rainbow_dqn.learn(experiences)
    clone_agent = rainbow_dqn.clone()

    assert clone_agent.observation_space == rainbow_dqn.observation_space
    assert clone_agent.action_space == rainbow_dqn.action_space
    # assert clone_agent.actor_network == rainbow_dqn.actor_network
    assert clone_agent.batch_size == rainbow_dqn.batch_size
    assert clone_agent.lr == rainbow_dqn.lr
    assert clone_agent.learn_step == rainbow_dqn.learn_step
    assert clone_agent.gamma == rainbow_dqn.gamma
    assert clone_agent.tau == rainbow_dqn.tau
    assert clone_agent.mut == rainbow_dqn.mut
    assert clone_agent.device == rainbow_dqn.device
    assert clone_agent.accelerator == rainbow_dqn.accelerator
    assert_state_dicts_equal(
        clone_agent.actor.state_dict(), rainbow_dqn.actor.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), rainbow_dqn.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), rainbow_dqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == rainbow_dqn.fitness
    assert clone_agent.steps == rainbow_dqn.steps
    assert clone_agent.scores == rainbow_dqn.scores
    rainbow_dqn.clean_up()
    clone_agent.clean_up()



import copy

import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces

from agilerl.algorithms.td3 import TD3
from agilerl.modules import EvolvableCNN, EvolvableMLP, EvolvableMultiInput
from agilerl.networks.actors import DeterministicActor
from agilerl.networks.q_networks import ContinuousQNetwork
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import (
    assert_not_equal_state_dict,
    assert_state_dicts_equal,
    get_experiences_batch,
    get_sample_from_space,
)


class DummyTD3(TD3):
    def __init__(self, observation_space, action_space, *args, **kwargs):
        super().__init__(observation_space, action_space, *args, **kwargs)

        self.tensor_test = torch.randn(1)


class DummyEnv:
    def __init__(self, state_size, vect=True, num_envs=2):
        self.state_size = state_size
        self.vect = vect
        if self.vect:
            self.state_size = (num_envs,) + self.state_size
            self.n_envs = num_envs
            self.num_envs = num_envs
        else:
            self.n_envs = 1

    def reset(self):
        return np.random.rand(*self.state_size), {}

    def step(self, action):
        return (
            np.random.rand(*self.state_size),
            np.random.randint(0, 5, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            {},
        )


class SimpleCNN(nn.Module):
    def __init__(self):
        super().__init__()

        self.conv1 = nn.Conv2d(
            3, 16, kernel_size=3, stride=1, padding=1
        )  # Input channels: 3 (for RGB images), Output channels: 16
        self.relu1 = nn.ReLU()
        self.mp1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv2d(
            16, 32, kernel_size=3, stride=1, padding=1
        )  # Input channels: 16, Output channels: 32
        self.relu2 = nn.ReLU()
        self.mp2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.flat = nn.Flatten()  # Flatten the 2D feature map to a 1D vector
        self.linear1 = nn.Linear(
            32 * 8 * 8, 128
        )  # Fully connected layer with 128 output features
        self.relu3 = nn.ReLU()
        self.linear2 = nn.Linear(
            128, 128
        )  # Fully connected layer with 128 output features

    def forward(self, x, xc):
        x = self.mp1(self.relu1(self.conv1(x)))
        x = self.mp2(self.relu2(self.conv2(x)))
        x = self.flat(x)
        x = self.relu3(self.linear1(x))
        x = self.relu3(self.linear2(x))
        return x


# initialize td3 with valid parameters
# Initializes all necessary attributes with default values
@pytest.mark.parametrize(
    "observation_space, encoder_cls",
    [
        ("vector_space", EvolvableMLP),
        ("image_space", EvolvableCNN),
        ("dict_space", EvolvableMultiInput),
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_initialize_td3(
    observation_space, vector_space, encoder_cls, accelerator, request
):
    observation_space = request.getfixturevalue(observation_space)

    # Initialize TD3 with default parameters
    td3 = TD3(observation_space, vector_space, accelerator=accelerator)

    expected_opt_cls = AcceleratedOptimizer if accelerator else optim.Adam
    expected_device = accelerator.device if accelerator else "cpu"
    assert td3.observation_space == observation_space
    assert td3.action_space == vector_space
    assert td3.batch_size == 64
    assert td3.lr_actor == 0.0001
    assert td3.lr_critic == 0.001
    assert td3.learn_step == 5
    assert td3.gamma == 0.99
    assert td3.tau == 0.005
    assert td3.mut is None
    assert td3.device == expected_device
    assert td3.accelerator == accelerator
    assert td3.index == 0
    assert td3.scores == []
    assert td3.fitness == []
    assert td3.steps == [0]
    assert isinstance(td3.actor.encoder, encoder_cls)
    assert isinstance(td3.actor_target.encoder, encoder_cls)
    assert isinstance(td3.actor_optimizer.optimizer, expected_opt_cls)
    assert isinstance(td3.critic_1.encoder, encoder_cls)
    assert isinstance(td3.critic_target_1.encoder, encoder_cls)
    assert isinstance(td3.critic_1_optimizer.optimizer, expected_opt_cls)
    assert isinstance(td3.critic_2.encoder, encoder_cls)
    assert isinstance(td3.critic_target_2.encoder, encoder_cls)
    assert isinstance(td3.critic_2_optimizer.optimizer, expected_opt_cls)
    assert isinstance(td3.criterion, nn.MSELoss)
    td3.clean_up()


# Can initialize td3 with an actor network
# TODO: This will be deprecated in the future
@pytest.mark.parametrize(
    "observation_space, actor_network, critic_1_network, critic_2_network, input_tensor, input_tensor_critic",
    [
        (
            "vector_space",
            "simple_mlp",
            "simple_mlp_critic",
            "simple_mlp_critic",
            torch.randn(1, 4),
            torch.randn(1, 6),
        ),
    ],
)
def test_initialize_td3_with_actor_network(
    observation_space,
    vector_space,
    actor_network,
    critic_1_network,
    critic_2_network,
    input_tensor,
    input_tensor_critic,
    request,
):
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)
    critic_1_network = request.getfixturevalue(critic_1_network)
    critic_1_network = MakeEvolvable(critic_1_network, input_tensor_critic)
    critic_2_network = request.getfixturevalue(critic_2_network)
    critic_2_network = MakeEvolvable(critic_2_network, input_tensor_critic)

    td3 = TD3(
        observation_space,
        vector_space,
        expl_noise=np.zeros((1, vector_space.shape[0])),
        actor_network=actor_network,
        critic_networks=[critic_1_network, critic_2_network],
    )

    assert td3.observation_space == observation_space
    assert td3.action_space == vector_space
    assert td3.batch_size == 64
    assert td3.lr_actor == 0.0001
    assert td3.lr_critic == 0.001
    assert td3.learn_step == 5
    assert td3.gamma == 0.99
    assert td3.tau == 0.005
    assert td3.mut is None
    assert td3.device == "cpu"
    assert td3.accelerator is None
    assert td3.index == 0
    assert td3.scores == []
    assert td3.fitness == []
    assert td3.steps == [0]
    assert isinstance(td3.actor_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.critic_1_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.critic_2_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.criterion, nn.MSELoss)
    td3.clean_up()


# Can initialize td3 with an actor network
# TODO: This will be deprecated in the future
@pytest.mark.parametrize(
    "observation_space, actor_network, critic_1_network, critic_2_network, input_tensor, input_tensor_critic",
    [
        (
            "vector_space",
            "simple_mlp",
            "simple_mlp_critic",
            "simple_mlp_critic",
            torch.randn(1, 4),
            torch.randn(1, 6),
        ),
    ],
)

# Can initialize td3 with an actor network but no critics - should trigger warning
# TODO: This will be deprecated in the future
def test_initialize_td3_with_actor_network_no_critics(
    observation_space,
    vector_space,
    actor_network,
    critic_1_network,
    critic_2_network,
    input_tensor,
    input_tensor_critic,
    request,
):
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)

    td3 = TD3(
        observation_space,
        vector_space,
        actor_network=actor_network,
        critic_networks=None,
    )

    assert td3.observation_space == observation_space
    assert td3.action_space == vector_space
    assert td3.batch_size == 64
    assert td3.lr_actor == 0.0001
    assert td3.lr_critic == 0.001
    assert td3.learn_step == 5
    assert td3.gamma == 0.99
    assert td3.tau == 0.005
    assert td3.mut is None
    assert td3.device == "cpu"
    assert td3.accelerator is None
    assert td3.index == 0
    assert td3.scores == []
    assert td3.fitness == []
    assert td3.steps == [0]
    assert isinstance(td3.actor_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.critic_1_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.critic_2_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.criterion, nn.MSELoss)
    td3.clean_up()


# Can initialize td3 with an actor network
# TODO: This will be deprecated in the future
@pytest.mark.parametrize(
    "observation_space, actor_network, input_tensor",
    [
        (
            "image_space",
            "simple_cnn",
            torch.randn(1, 3, 32, 32),
        ),
    ],
)
def test_initialize_td3_with_actor_network_cnn(
    observation_space, vector_space, actor_network, input_tensor, request
):
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)
    critic_1_network = SimpleCNN()
    critic_1_network = MakeEvolvable(
        critic_1_network,
        input_tensor,
        torch.randn(1, vector_space.shape[0]),
    )
    critic_2_network = SimpleCNN()
    critic_2_network = MakeEvolvable(
        critic_2_network,
        input_tensor,
        torch.randn(1, vector_space.shape[0]),
    )

    td3 = TD3(
        observation_space,
        vector_space,
        actor_network=actor_network,
        critic_networks=[critic_1_network, critic_2_network],
    )

    assert td3.observation_space == observation_space
    assert td3.action_space == vector_space
    assert td3.batch_size == 64
    assert td3.lr_actor == 0.0001
    assert td3.lr_critic == 0.001
    assert td3.learn_step == 5
    assert td3.gamma == 0.99
    assert td3.tau == 0.005
    assert td3.mut is None
    assert td3.device == "cpu"
    assert td3.accelerator is None
    assert td3.index == 0
    assert td3.scores == []
    assert td3.fitness == []
    assert td3.steps == [0]
    assert isinstance(td3.actor_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.critic_1_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.critic_2_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.criterion, nn.MSELoss)
    td3.clean_up()


@pytest.mark.parametrize(
    "observation_space", ["vector_space", "image_space", "dict_space"]
)
def test_returns_expected_action_training(observation_space, request):
    observation_space = request.getfixturevalue(observation_space)
    action_space = spaces.Box(low=-1, high=1, shape=(2,))
    accelerator = Accelerator()

    td3 = TD3(observation_space, action_space)
    state = get_sample_from_space(observation_space)
    training = False
    action = td3.get_action(state, training)[0]

    assert len(action) == action_space.shape[0]
    for act in action:
        assert isinstance(act, np.float32)
        assert -1 <= act <= 1
    td3.clean_up()

    td3 = TD3(
        observation_space,
        action_space,
        accelerator=accelerator,
    )
    state = get_sample_from_space(observation_space)
    training = True
    action = td3.get_action(state, training)[0]

    assert len(action) == action_space.shape[0]
    for act in action:
        assert isinstance(act, np.float32)
        assert -1 <= act <= 1
    td3.clean_up()

    td3 = TD3(
        observation_space,
        action_space,
        O_U_noise=False,
        accelerator=accelerator,
    )
    state = get_sample_from_space(observation_space)
    training = True
    action = td3.get_action(state, training)[0]

    assert len(action) == action_space.shape[0]
    for act in action:
        assert isinstance(act, np.float32)
        assert -1 <= act <= 1
    td3.clean_up()


# Returns the expected action from float64 input
def test_returns_expected_action_float64(discrete_space):
    action_space = spaces.Box(low=-1, high=1, shape=(2,))
    td3 = TD3(discrete_space, action_space)
    state = np.array([0, 1, 0, 1]).astype(np.float64)
    training = False
    action = td3.get_action(state, training)[0]

    assert len(action) == action_space.shape[0]
    for act in action:
        assert isinstance(act, np.float32)
        assert -1 <= act <= 1
    td3.clean_up()

    td3 = TD3(discrete_space, action_space)
    state = np.array([1]).astype(np.float64)
    training = True
    action = td3.get_action(state, training)[0]

    assert len(action) == action_space.shape[0]
    for act in action:
        assert isinstance(act, np.float32)
        assert -1 <= act <= 1
    td3.clean_up()

    td3 = TD3(discrete_space, action_space, O_U_noise=False)
    state = np.array([1]).astype(np.float64)
    training = True
    action = td3.get_action(state, training)[0]

    assert len(action) == action_space.shape[0]
    for act in action:
        assert isinstance(act, np.float32)
        assert -1 <= act <= 1
    td3.clean_up()


# learns from experiences and updates network parameters
@pytest.mark.parametrize(
    "observation_space", ["vector_space", "image_space", "dict_space"]
)
@pytest.mark.parametrize(
    "min_action, max_action",
    [(-1, 1), ([-1, 0], [1, 1]), ([-1, -1], [0, 1]), ([-1, -2], [1, 0])],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_learns_from_experiences(
    observation_space, min_action, max_action, accelerator, request
):
    observation_space = request.getfixturevalue(observation_space)
    # Continuous action space
    low = np.array(min_action) if isinstance(min_action, list) else min_action
    high = np.array(max_action) if isinstance(max_action, list) else max_action
    action_space = spaces.Box(low=low, high=high, shape=(2,))

    # Create an instance of the td3 class
    batch_size = 64
    policy_freq = 2
    td3 = TD3(
        observation_space,
        action_space,
        batch_size=batch_size,
        policy_freq=policy_freq,
        accelerator=accelerator,
    )

    # Create a batch of experiences
    device = accelerator.device if accelerator else "cpu"
    experiences = get_experiences_batch(
        observation_space, action_space, batch_size, device
    )

    # Copy state dict before learning - should be different to after updating weights
    actor = td3.actor
    actor_target = td3.actor_target
    actor_pre_learn_sd = copy.deepcopy(td3.actor.state_dict())
    critic_1 = td3.critic_1
    critic_target_1 = td3.critic_target_1
    critic_1_pre_learn_sd = copy.deepcopy(td3.critic_1.state_dict())
    critic_2 = td3.critic_2
    critic_target_2 = td3.critic_target_2
    critic_2_pre_learn_sd = copy.deepcopy(td3.critic_2.state_dict())

    td3.scores = [0]

    actor_loss, critic_loss = td3.learn(experiences)
    assert actor_loss is None
    assert isinstance(critic_loss, float)
    assert critic_loss >= 0.0

    td3.scores = [0, 0]
    # Call the learn method
    actor_loss, critic_loss = td3.learn(experiences)

    assert isinstance(actor_loss, float)
    assert isinstance(critic_loss, float)
    assert critic_loss >= 0.0
    assert actor == td3.actor
    assert actor_target == td3.actor_target
    assert_not_equal_state_dict(actor_pre_learn_sd, td3.actor.state_dict())
    assert critic_1 == td3.critic_1
    assert critic_target_1 == td3.critic_target_1
    assert_not_equal_state_dict(critic_1_pre_learn_sd, td3.critic_1.state_dict())
    assert critic_2 == td3.critic_2
    assert critic_target_2 == td3.critic_target_2
    assert_not_equal_state_dict(critic_2_pre_learn_sd, td3.critic_2.state_dict())
    td3.clean_up()


# Updates target network parameters with soft update
def test_soft_update(vector_space):
    net_config = {"encoder_config": {"hidden_size": [64, 64]}}
    batch_size = 64
    lr_actor = 1e-4
    lr_critic = 1e-3
    learn_step = 5
    gamma = 0.99
    tau = 1e-3
    mut = None
    actor_network = None
    device = "cpu"
    accelerator = None
    wrap = True

    td3 = TD3(
        vector_space,
        copy.deepcopy(vector_space),
        net_config=net_config,
        batch_size=batch_size,
        lr_actor=lr_actor,
        lr_critic=lr_critic,
        learn_step=learn_step,
        gamma=gamma,
        tau=tau,
        mut=mut,
        actor_network=actor_network,
        device=device,
        accelerator=accelerator,
        wrap=wrap,
    )

    td3.soft_update(td3.actor, td3.actor_target)

    eval_params = list(td3.actor.parameters())
    target_params = list(td3.actor_target.parameters())
    expected_params = [
        td3.tau * eval_param + (1.0 - td3.tau) * target_param
        for eval_param, target_param in zip(eval_params, target_params)
    ]

    assert all(
        torch.allclose(expected_param, target_param)
        for expected_param, target_param in zip(expected_params, target_params)
    )

    td3.soft_update(td3.critic_1, td3.critic_target_1)

    eval_params = list(td3.critic_1.parameters())
    target_params = list(td3.critic_target_1.parameters())
    expected_params = [
        td3.tau * eval_param + (1.0 - td3.tau) * target_param
        for eval_param, target_param in zip(eval_params, target_params)
    ]

    assert all(
        torch.allclose(expected_param, target_param)
        for expected_param, target_param in zip(expected_params, target_params)
    )

    td3.soft_update(td3.critic_2, td3.critic_target_2)

    eval_params = list(td3.critic_2.parameters())
    target_params = list(td3.critic_target_2.parameters())
    expected_params = [
        td3.tau * eval_param + (1.0 - td3.tau) * target_param
        for eval_param, target_param in zip(eval_params, target_params)
    ]

    assert all(
        torch.allclose(expected_param, target_param)
        for expected_param, target_param in zip(expected_params, target_params)
    )
    td3.clean_up()


# Runs algorithm test loop
@pytest.mark.parametrize("observation_space", ["vector_space", "image_space"])
@pytest.mark.parametrize("num_envs", [1, 3])
def test_algorithm_test_loop(observation_space, vector_space, num_envs, request):
    observation_space = request.getfixturevalue(observation_space)
    vect = num_envs > 1
    env = DummyEnv(state_size=observation_space.shape, vect=vect, num_envs=num_envs)
    agent = TD3(observation_space, vector_space)
    mean_score = agent.test(env, max_steps=10)
    assert isinstance(mean_score, float)
    agent.clean_up()


# Clones the agent and returns an identical agent.
@pytest.mark.parametrize("observation_space", ["vector_space"])
def test_clone_returns_identical_agent(observation_space, vector_space, request):
    observation_space = request.getfixturevalue(observation_space)

    td3 = DummyTD3(observation_space, vector_space)
    td3.fitness = [200, 200, 200]
    td3.scores = [94, 94, 94]
    td3.steps = [2500]
    td3.tensor_attribute = torch.randn(1)
    clone_agent = td3.clone()

    assert clone_agent.observation_space == td3.observation_space
    assert clone_agent.action_space == td3.action_space
    assert clone_agent.batch_size == td3.batch_size
    assert clone_agent.lr_actor == td3.lr_actor
    assert clone_agent.lr_critic == td3.lr_critic
    assert clone_agent.learn_step == td3.learn_step
    assert clone_agent.gamma == td3.gamma
    assert clone_agent.tau == td3.tau
    assert clone_agent.mut == td3.mut
    assert clone_agent.device == td3.device
    assert clone_agent.accelerator == td3.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), td3.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), td3.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_1.state_dict(), td3.critic_1.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_target_1.state_dict(), td3.critic_target_1.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_2.state_dict(), td3.critic_2.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_target_2.state_dict(), td3.critic_target_2.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.actor_optimizer.state_dict(), td3.actor_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_1_optimizer.state_dict(), td3.critic_1_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_2_optimizer.state_dict(), td3.critic_2_optimizer.state_dict()
    )
    assert clone_agent.fitness == td3.fitness
    assert clone_agent.steps == td3.steps
    assert clone_agent.scores == td3.scores
    assert clone_agent.tensor_attribute == td3.tensor_attribute
    assert clone_agent.tensor_test == td3.tensor_test
    td3.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    td3 = TD3(observation_space, vector_space, accelerator=accelerator)
    clone_agent = td3.clone()

    assert clone_agent.observation_space == td3.observation_space
    assert clone_agent.action_space == td3.action_space
    assert clone_agent.batch_size == td3.batch_size
    assert clone_agent.lr_actor == td3.lr_actor
    assert clone_agent.lr_critic == td3.lr_critic
    assert clone_agent.learn_step == td3.learn_step
    assert clone_agent.gamma == td3.gamma
    assert clone_agent.tau == td3.tau
    assert clone_agent.mut == td3.mut
    assert clone_agent.device == td3.device
    assert clone_agent.accelerator == td3.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), td3.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), td3.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_1.state_dict(), td3.critic_1.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_target_1.state_dict(), td3.critic_target_1.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_2.state_dict(), td3.critic_2.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_target_2.state_dict(), td3.critic_target_2.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.actor_optimizer.state_dict(), td3.actor_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_1_optimizer.state_dict(), td3.critic_1_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_2_optimizer.state_dict(), td3.critic_2_optimizer.state_dict()
    )
    assert clone_agent.fitness == td3.fitness
    assert clone_agent.steps == td3.steps
    assert clone_agent.scores == td3.scores
    td3.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    td3 = TD3(observation_space, vector_space, accelerator=accelerator, wrap=False)
    clone_agent = td3.clone(wrap=False)

    assert clone_agent.observation_space == td3.observation_space
    assert clone_agent.action_space == td3.action_space
    assert clone_agent.batch_size == td3.batch_size
    assert clone_agent.lr_actor == td3.lr_actor
    assert clone_agent.lr_critic == td3.lr_critic
    assert clone_agent.learn_step == td3.learn_step
    assert clone_agent.gamma == td3.gamma
    assert clone_agent.tau == td3.tau
    assert clone_agent.mut == td3.mut
    assert clone_agent.device == td3.device
    assert clone_agent.accelerator == td3.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), td3.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), td3.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_1.state_dict(), td3.critic_1.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_target_1.state_dict(), td3.critic_target_1.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_2.state_dict(), td3.critic_2.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_target_2.state_dict(), td3.critic_target_2.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.actor_optimizer.state_dict(), td3.actor_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_1_optimizer.state_dict(), td3.critic_1_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_2_optimizer.state_dict(), td3.critic_2_optimizer.state_dict()
    )
    assert clone_agent.fitness == td3.fitness
    assert clone_agent.steps == td3.steps
    assert clone_agent.scores == td3.scores
    td3.clean_up()
    clone_agent.clean_up()


def test_clone_new_index(vector_space):
    td3 = TD3(vector_space, copy.deepcopy(vector_space))
    clone_agent = td3.clone(index=100)

    assert clone_agent.index == 100
    td3.clean_up()
    clone_agent.clean_up()


def test_clone_after_learning(vector_space):
    batch_size = 8
    td3 = TD3(vector_space, copy.deepcopy(vector_space))

    # Get experiences and learn
    experiences = get_experiences_batch(
        vector_space, vector_space, batch_size, device=td3.device
    )
    td3.learn(experiences)

    # Clone the agent
    clone_agent = td3.clone()

    assert clone_agent.observation_space == td3.observation_space
    assert clone_agent.action_space == td3.action_space
    assert clone_agent.batch_size == td3.batch_size
    assert clone_agent.lr_actor == td3.lr_actor
    assert clone_agent.lr_critic == td3.lr_critic
    assert clone_agent.learn_step == td3.learn_step
    assert clone_agent.gamma == td3.gamma
    assert clone_agent.tau == td3.tau
    assert clone_agent.mut == td3.mut
    assert clone_agent.device == td3.device
    assert clone_agent.accelerator == td3.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), td3.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), td3.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_1.state_dict(), td3.critic_1.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_target_1.state_dict(), td3.critic_target_1.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_2.state_dict(), td3.critic_2.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_target_2.state_dict(), td3.critic_target_2.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.actor_optimizer.state_dict(), td3.actor_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_1_optimizer.state_dict(), td3.critic_1_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_2_optimizer.state_dict(), td3.critic_2_optimizer.state_dict()
    )
    assert clone_agent.fitness == td3.fitness
    assert clone_agent.steps == td3.steps
    assert clone_agent.scores == td3.scores
    td3.clean_up()
    clone_agent.clean_up()


@pytest.mark.parametrize("observation_space", ["vector_space"])
def test_initialize_td3_with_actor_network_evo_net(
    observation_space, vector_space, request
):
    observation_space = request.getfixturevalue(observation_space)

    actor_network = DeterministicActor(observation_space, vector_space)
    critic_networks = [
        ContinuousQNetwork(observation_space, vector_space) for _ in range(2)
    ]

    td3 = TD3(
        observation_space,
        vector_space,
        actor_network=actor_network,
        critic_networks=critic_networks,
    )

    assert td3.observation_space == observation_space
    assert td3.action_space == vector_space
    assert td3.batch_size == 64
    assert td3.lr_actor == 0.0001
    assert td3.lr_critic == 0.001
    assert td3.learn_step == 5
    assert td3.gamma == 0.99
    assert td3.tau == 0.005
    assert td3.mut is None
    assert td3.device == "cpu"
    assert td3.accelerator is None
    assert td3.index == 0
    assert td3.scores == []
    assert td3.fitness == []
    assert td3.steps == [0]
    assert isinstance(td3.actor_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.critic_1_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.critic_2_optimizer.optimizer, optim.Adam)
    assert isinstance(td3.criterion, nn.MSELoss)
    td3.clean_up()


def test_initialize_td3_with_incorrect_actor_net(vector_space):
    actor_network = "dummy"
    critic_networks = "dummy"
    with pytest.raises(AssertionError):
        td3 = TD3(
            vector_space,
            copy.deepcopy(vector_space),
            actor_network=actor_network,
            critic_networks=critic_networks,
        )
        assert td3


@pytest.mark.parametrize(
    "min, max, action, expected_result, device",
    [
        (0, 1, [1.1, 0.75, -1], [1.0, 0.75, 0], "cpu"),
        ([0.5, 0, 0.1], 1, [0, 0, 0.2], [0.5, 0, 0.2], "cpu"),
        (0, [0.75, 1.0, 0.1], [1.0, 0.75, 0.1], [0.75, 0.75, 0.1], "cpu"),
        (
            [-1, -1, -1],
            [1, 1, 1],
            [[-2, 1, 0.25], [1.5, -1, 0.75]],
            [[-1, 1, 0.25], [1, -1, 0.75]],
            "cpu",
        ),
        (0, 1, [1.1, 0.75, -1], [1.0, 0.75, 0], "cuda"),
        ([0.5, 0, 0.1], 1, [0, 0, 0.2], [0.5, 0, 0.2], "cuda"),
        (0, [0.75, 1.0, 0.1], [1.0, 0.75, 0.1], [0.75, 0.75, 0.1], "cuda"),
        (
            [-1, -1, -1],
            [1, 1, 1],
            [[-2, 1, 0.25], [1.5, -1, 0.75]],
            [[-1, 1, 0.25], [1, -1, 0.75]],
            "cuda",
        ),
    ],
)
def test_multi_dim_clamp(vector_space, min, max, action, expected_result, device):
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA not available")

    if isinstance(min, list):
        min = np.array(min)
    if isinstance(max, list):
        max = np.array(max)
    td3 = TD3(
        vector_space,
        copy.deepcopy(vector_space),
        device=device,
    )
    input = torch.tensor(action, dtype=torch.float32).to(device)
    clamped_actions = td3.multi_dim_clamp(min, max, input).type(torch.float32)
    expected_result = torch.tensor(expected_result)
    assert clamped_actions.dtype == expected_result.dtype
    td3.clean_up()



import copy
import warnings
from typing import Any, Callable, Optional, Union

import numpy as np
import torch
import torch.optim as optim
from gymnasium import spaces
from tensordict import TensorDict
from torch.nn.utils import clip_grad_norm_

from agilerl.algorithms.core import OptimizerWrapper, RLAlgorithm
from agilerl.algorithms.core.registry import HyperparameterConfig, NetworkGroup
from agilerl.components.rollout_buffer import RolloutBuffer
from agilerl.modules.base import EvolvableModule
from agilerl.modules.configs import MlpNetConfig
from agilerl.networks import EvolvableNetwork, StochasticActor
from agilerl.networks.value_networks import ValueNetwork
from agilerl.typing import ArrayOrTensor, BPTTSequenceType, ExperiencesType, GymEnvType
from agilerl.utils.algo_utils import (
    flatten_experiences,
    get_experiences_samples,
    is_vectorized_experiences,
    make_safe_deepcopies,
    obs_channels_to_first,
    share_encoder_parameters,
    stack_experiences,
)


class PPO(RLAlgorithm):
    """Proximal Policy Optimization (PPO) algorithm.

    Paper: https://arxiv.org/abs/1707.06347v2

    :param observation_space: Observation space of the environment
    :type observation_space: gym.spaces.Space
    :param action_space: Action space of the environment
    :type action_space: gym.spaces.Space
    :param index: Index to keep track of object instance during tournament selection and mutation, defaults to 0
    :type index: int, optional
    :param hp_config: RL hyperparameter mutation configuration, defaults to None, whereby algorithm mutations are disabled.
    :type hp_config: HyperparameterConfig, optional
    :param net_config: Network configuration, defaults to None
    :type net_config: dict, optional
    :param batch_size: Size of batched sample from replay buffer for learning, defaults to 64
    :type batch_size: int, optional
    :param lr: Learning rate for optimizer, defaults to 1e-4
    :type lr: float, optional
    :param learn_step: Learning frequency, defaults to 2048
    :type learn_step: int, optional
    :param gamma: Discount factor, defaults to 0.99
    :type gamma: float, optional
    :param gae_lambda: Lambda for general advantage estimation, defaults to 0.95
    :type gae_lambda: float, optional
    :param mut: Most recent mutation to agent, defaults to None
    :type mut: str, optional
    :param action_std_init: Initial action standard deviation, defaults to 0.0
    :type action_std_init: float, optional
    :param clip_coef: Surrogate clipping coefficient, defaults to 0.2
    :type clip_coef: float, optional
    :param ent_coef: Entropy coefficient, defaults to 0.01
    :type ent_coef: float, optional
    :param vf_coef: Value function coefficient, defaults to 0.5
    :type vf_coef: float, optional
    :param max_grad_norm: Maximum norm for gradient clipping, defaults to 0.5
    :type max_grad_norm: float, optional
    :param target_kl: Target KL divergence threshold, defaults to None
    :type target_kl: float, optional
    :param normalize_images: Flag to normalize images, defaults to True
    :type normalize_images: bool, optional
    :param update_epochs: Number of policy update epochs, defaults to 4
    :type update_epochs: int, optional
    :param actor_network: Custom actor network, defaults to None
    :type actor_network: nn.Module, optional
    :param critic_network: Custom critic network, defaults to None
    :type critic_network: nn.Module, optional
    :param share_encoders: Flag to share encoder parameters between actor and critic, defaults to False
    :type share_encoders: bool, optional
    :param num_envs: Number of parallel environments, defaults to 1
    :type num_envs: int, optional
    :param use_rollout_buffer: Flag to use the rollout buffer instead of tuple experiences, defaults to False
    :type use_rollout_buffer: bool, optional
    :param recurrent: Flag to use hidden states for recurrent policies, defaults to False
    :type recurrent: bool, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wrap: Wrap models for distributed training upon creation, defaults to True
    :type wrap: bool, optional
    :param bptt_sequence_type: Type of sequence for BPTT learning, defaults to BPTTSequenceType.CHUNKED
    :type bptt_sequence_type: BPTTSequenceType, optional
    :param max_seq_len: Maximum sequence length for truncated BPTT, defaults to None, where complete episodes
    are used as sequences.
    :type max_seq_len: int, optional
    """

    def __init__(
        self,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        index: int = 0,
        hp_config: Optional[HyperparameterConfig] = None,
        net_config: Optional[dict[str, Any]] = None,
        batch_size: int = 64,
        lr: float = 1e-4,
        learn_step: int = 2048,
        gamma: float = 0.99,
        gae_lambda: float = 0.95,
        mut: Optional[str] = None,
        action_std_init: float = 0.0,
        clip_coef: float = 0.2,
        ent_coef: float = 0.01,
        vf_coef: float = 0.5,
        max_grad_norm: float = 0.5,
        target_kl: Optional[float] = None,
        normalize_images: bool = True,
        update_epochs: int = 4,
        actor_network: Optional[EvolvableModule] = None,
        critic_network: Optional[EvolvableModule] = None,
        share_encoders: bool = True,
        num_envs: int = 1,
        use_rollout_buffer: bool = False,
        rollout_buffer_config: Optional[dict[str, Any]] = {},
        recurrent: bool = False,
        device: str = "cpu",
        accelerator: Optional[Any] = None,
        wrap: bool = True,
        bptt_sequence_type: BPTTSequenceType = BPTTSequenceType.CHUNKED,
        max_seq_len: Optional[int] = None,
    ) -> None:
        super().__init__(
            observation_space,
            action_space,
            index=index,
            hp_config=hp_config,
            device=device,
            accelerator=accelerator,
            normalize_images=normalize_images,
            name="PPO",
        )

        assert learn_step >= 1, "Learn step must be greater than or equal to one."
        assert isinstance(learn_step, int), "Learn step must be an integer."
        assert isinstance(batch_size, int), "Batch size must be an integer."
        assert batch_size >= 1, "Batch size must be greater than or equal to one."
        assert isinstance(lr, float), "Learning rate must be a float."
        assert lr > 0, "Learning rate must be greater than zero."
        assert isinstance(gamma, (float, int, torch.Tensor)), "Gamma must be a float."
        assert isinstance(gae_lambda, (float, int)), "Lambda must be a float."
        assert gae_lambda >= 0, "Lambda must be greater than or equal to zero."
        assert isinstance(
            action_std_init, (float, int)
        ), "Action standard deviation must be a float."
        assert (
            action_std_init >= 0
        ), "Action standard deviation must be greater than or equal to zero."
        assert isinstance(
            clip_coef, (float, int)
        ), "Clipping coefficient must be a float."
        assert (
            clip_coef >= 0
        ), "Clipping coefficient must be greater than or equal to zero."
        assert isinstance(
            ent_coef, (float, int)
        ), "Entropy coefficient must be a float."
        assert (
            ent_coef >= 0
        ), "Entropy coefficient must be greater than or equal to zero."
        assert isinstance(
            vf_coef, (float, int)
        ), "Value function coefficient must be a float."
        assert (
            vf_coef >= 0
        ), "Value function coefficient must be greater than or equal to zero."
        assert isinstance(
            max_grad_norm, (float, int)
        ), "Maximum norm for gradient clipping must be a float."
        assert (
            max_grad_norm >= 0
        ), "Maximum norm for gradient clipping must be greater than or equal to zero."
        assert (
            isinstance(target_kl, (float, int)) or target_kl is None
        ), "Target KL divergence threshold must be a float."
        if target_kl is not None:
            assert (
                target_kl >= 0
            ), "Target KL divergence threshold must be greater than or equal to zero."
        assert isinstance(
            update_epochs, int
        ), "Policy update epochs must be an integer."
        assert (
            update_epochs >= 1
        ), "Policy update epochs must be greater than or equal to one."
        assert isinstance(
            wrap, bool
        ), "Wrap models flag must be boolean value True or False."

        # New parameters for using RolloutBuffer
        assert isinstance(
            use_rollout_buffer, bool
        ), "Use rollout buffer flag must be boolean value True or False."
        assert isinstance(
            recurrent, bool
        ), "Has hidden states flag must be boolean value True or False."
        assert isinstance(
            bptt_sequence_type, BPTTSequenceType
        ), "bptt_sequence_type must be a BPTTSequenceType enum value."

        if not use_rollout_buffer:
            warnings.warn(
                (
                    "DeprecationWarning: 'use_rollout_buffer=False' is deprecated and will be removed in a future release. "
                    "The PPO implementation now expects 'use_rollout_buffer=True' for improved performance, "
                    "cleaner support for recurrent policies, and easier integration with custom environments. "
                    "Please update your code to use 'use_rollout_buffer=True' and, if you require recurrent policies, set 'recurrent=True'.\n"
                    "Refer to the documentation for migration instructions and further details."
                ),
                DeprecationWarning,
                stacklevel=2,
            )

        self.recurrent = recurrent
        self.use_rollout_buffer = use_rollout_buffer
        self.net_config = net_config

        if self.recurrent and not self.use_rollout_buffer:
            raise ValueError("use_rollout_buffer must be True if recurrent=True.")

        self.max_seq_len = max_seq_len
        self.batch_size = batch_size
        self.lr = lr
        self.gamma = gamma
        self.learn_step = learn_step
        self.mut = mut
        self.gae_lambda = gae_lambda
        self.action_std_init = action_std_init
        self.clip_coef = clip_coef
        self.ent_coef = ent_coef
        self.vf_coef = vf_coef
        self.max_grad_norm = max_grad_norm
        self.target_kl = target_kl
        self.update_epochs = update_epochs
        self.num_envs = num_envs
        self.rollout_buffer_config = rollout_buffer_config
        self.bptt_sequence_type = bptt_sequence_type

        if actor_network is not None and critic_network is not None:
            if not isinstance(actor_network, EvolvableModule):
                raise TypeError(
                    f"Passed actor network is of type {type(actor_network)}, but must be of type EvolvableModule."
                )
            if not isinstance(critic_network, EvolvableModule):
                raise TypeError(
                    f"Passed critic network is of type {type(critic_network)}, but must be of type EvolvableModule."
                )

            self.actor, self.critic = make_safe_deepcopies(
                actor_network, critic_network
            )
        else:
            net_config_dict = {} if self.net_config is None else self.net_config

            critic_net_config = copy.deepcopy(net_config_dict)

            head_config = net_config_dict.get("head_config", None)
            if head_config is not None:
                critic_head_config = copy.deepcopy(head_config)
                critic_head_config["output_activation"] = None
                critic_net_config.pop("squash_output", None)
            else:
                critic_head_config = MlpNetConfig(hidden_size=[16])

            critic_net_config["head_config"] = critic_head_config

            self.actor = StochasticActor(
                self.observation_space,
                self.action_space,
                action_std_init=self.action_std_init,
                device=self.device,
                recurrent=self.recurrent,
                encoder_name=("shared_encoder" if share_encoders else "actor_encoder"),
                **net_config_dict,
            )

            self.critic = ValueNetwork(
                self.observation_space,
                device=self.device,
                recurrent=self.recurrent,
                encoder_name=("shared_encoder" if share_encoders else "critic_encoder"),
                **critic_net_config,
            )

        # Share encoders between actor and critic
        self.share_encoders = share_encoders
        if self.share_encoders and all(
            isinstance(net, EvolvableNetwork) for net in [self.actor, self.critic]
        ):
            self.share_encoder_parameters()
            # Need to register a mutation hook that does this after every mutation
            self.register_mutation_hook(self.share_encoder_parameters)

        self.optimizer = OptimizerWrapper(
            optim.Adam, networks=[self.actor, self.critic], lr=self.lr
        )

        # Initialize rollout buffer if enabled
        if self.use_rollout_buffer:
            self.create_rollout_buffer()
            # Need to register a mutation hook that does this after every mutation (e.g. the batch size, sequence length, etc. have changed)
            self.register_mutation_hook(self.create_rollout_buffer)

        if self.accelerator is not None and wrap:
            self.wrap_models()

        # Register network groups for mutations
        self.register_network_group(NetworkGroup(eval_network=self.actor, policy=True))
        self.register_network_group(NetworkGroup(eval_network=self.critic))

        self.hidden_state = None

    def share_encoder_parameters(self) -> None:
        """Shares the encoder parameters between the actor and critic."""
        if all(isinstance(net, EvolvableNetwork) for net in [self.actor, self.critic]):
            share_encoder_parameters(self.actor, self.critic)
        else:
            warnings.warn(
                "Encoder sharing is disabled as actor or critic is not an EvolvableNetwork."
            )

    def create_rollout_buffer(self) -> None:
        """Creates a rollout buffer with the current configuration."""
        self.rollout_buffer = RolloutBuffer(
            capacity=-(self.learn_step // -self.num_envs),
            observation_space=self.observation_space,
            action_space=self.action_space,
            device=self.device,
            num_envs=self.num_envs,
            gae_lambda=self.gae_lambda,
            gamma=self.gamma,
            recurrent=self.recurrent,
            # recurrent specific parameters
            hidden_state_architecture=(
                self.get_hidden_state_architecture() if self.recurrent else None
            ),
            max_seq_len=self.max_seq_len if self.recurrent else None,
            bptt_sequence_type=self.bptt_sequence_type,
            **self.rollout_buffer_config,
        )

    def _extract_hidden_state(
        self, full_hidden_state: dict[str, ArrayOrTensor], encoder_name: str
    ) -> dict[str, ArrayOrTensor]:
        """Extract hidden state components for a specific network encoder.

        :param full_hidden_state: Complete hidden state dictionary
        :type full_hidden_state: dict[str, ArrayOrTensor]
        :param encoder_name: Name of the encoder to extract hidden states for
        :type encoder_name: str
        :return: Hidden state dictionary for the specific encoder
        :rtype: dict[str, ArrayOrTensor]
        """
        network_hidden_state = {}
        for key, value in full_hidden_state.items():
            if key.startswith(encoder_name):
                network_hidden_state[key] = value
        return network_hidden_state

    def _get_action_and_values(
        self,
        obs: ArrayOrTensor,
        action_mask: Optional[ArrayOrTensor] = None,
        hidden_state: Optional[
            dict[str, ArrayOrTensor]
        ] = None,  # Hidden state is a dict for recurrent policies
        *,
        sample: bool = True,
    ) -> tuple[
        ArrayOrTensor,
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
        Optional[dict[str, ArrayOrTensor]],
    ]:
        """
        Returns the next action to take in the environment and the values.

        :param obs: Environment observation, or multiple observations in a batch
        :type obs: ArrayOrTensor
        :param action_mask: Mask of legal actions 1=legal 0=illegal, defaults to None
        :type action_mask: Optional[ArrayOrTensor]
        :param hidden_state: Hidden state for recurrent policies, defaults to None
        :type hidden_state: Optional[dict[str, ArrayOrTensor]]
        :param sample: Whether to sample an action, defaults to True
        :type sample: bool
        :return: Action, log probability, entropy, state values, and (if recurrent) next hidden state
        :rtype: tuple[ArrayOrTensor, torch.Tensor, torch.Tensor, torch.Tensor, Optional[dict[str, ArrayOrTensor]]]
        """
        if hidden_state is not None:
            if self.share_encoders:
                # When sharing encoders, both networks use the same hidden state
                latent_pi, next_hidden_actor = self.actor.extract_features(
                    obs, hidden_state=hidden_state
                )
                action, log_prob, entropy = self.actor.forward_head(
                    latent_pi, action_mask=action_mask, sample=sample
                )
                values = self.critic.forward_head(latent_pi).squeeze(-1)
                next_hidden_combined = next_hidden_actor
            else:
                # When not sharing encoders, extract separate hidden states for actor and critic
                actor_hidden_state = self._extract_hidden_state(
                    hidden_state, "actor_encoder"
                )
                critic_hidden_state = self._extract_hidden_state(
                    hidden_state, "critic_encoder"
                )

                # Forward pass through actor with its hidden state
                latent_pi, next_hidden_actor = self.actor.extract_features(
                    obs, hidden_state=actor_hidden_state
                )
                action, log_prob, entropy = self.actor.forward_head(
                    latent_pi, action_mask=action_mask, sample=sample
                )

                # Forward pass through critic with its hidden state
                values, next_hidden_critic = self.critic(
                    obs, hidden_state=critic_hidden_state
                )
                values = values.squeeze(-1)

                # Combine the next hidden states from both networks
                next_hidden_combined = {}
                if next_hidden_actor is not None:
                    next_hidden_combined.update(next_hidden_actor)
                if next_hidden_critic is not None:
                    next_hidden_combined.update(next_hidden_critic)

            return action, log_prob, entropy, values, next_hidden_combined
        else:
            latent_pi = self.actor.extract_features(obs)
            action, log_prob, entropy = self.actor.forward_head(
                latent_pi, action_mask=action_mask, sample=sample
            )
            values = (
                self.critic.forward_head(latent_pi).squeeze(-1)
                if self.share_encoders
                else self.critic(obs).squeeze(-1)
            )
            return action, log_prob, entropy, values, None

    def get_hidden_state_architecture(self) -> dict[str, tuple[int, ...]]:
        """Get the hidden state architecture for the environment.

        :return: Dictionary describing the hidden state architecture (name to shape)
        :rtype: dict[str, tuple[int, ...]]
        """
        return {
            k: v.shape for k, v in self.get_initial_hidden_state(self.num_envs).items()
        }

    def get_initial_hidden_state(self, num_envs: int = 1) -> dict[str, ArrayOrTensor]:
        """Get the initial hidden state for the environment.

        The hidden states are generally cached on a per Module basis.
        The reason the Cache is per Module is because the user might want to have a custom initialization for the hidden states.

        :param num_envs: Number of environments, defaults to 1
        :type num_envs: int, optional
        :return: Initial hidden state dictionary
        :rtype: dict[str, ArrayOrTensor]
        """
        # Return a batch of initial hidden states
        # Flat map them into "actor_*" and "critic_*" (if not sharing encoders)
        flat_hidden = {}

        actor_hidden = self.actor.initialize_hidden_state(batch_size=num_envs)
        flat_hidden.update(actor_hidden)

        # also add the critic hidden state if not sharing encoders
        if not self.share_encoders:
            critic_hidden = self.critic.initialize_hidden_state(batch_size=num_envs)
            flat_hidden.update(critic_hidden)

        return flat_hidden

    def evaluate_actions(
        self,
        obs: ArrayOrTensor,
        actions: ArrayOrTensor,
        hidden_state: Optional[dict[str, ArrayOrTensor]] = None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Evaluates the actions.

        :param obs: Environment observation, or multiple observations in a batch
        :type obs: ArrayOrTensor
        :param actions: Actions to evaluate
        :type actions: ArrayOrTensor
        :param hidden_state: Hidden state for recurrent policies, defaults to None. Expected shape: dict with tensors of shape (batch_size, 1, hidden_size).
        :type hidden_state: Optional[dict[str, ArrayOrTensor]]
        :return: Log probability, entropy, state values
        :rtype: tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        """
        obs = self.preprocess_observation(obs)

        # Get values from actor-critic
        _, _, entropy, values, _ = self._get_action_and_values(
            obs, hidden_state=hidden_state, sample=False
        )

        log_prob = self.actor.action_log_prob(actions)

        # Use -log_prob as entropy when squashing output in continuous action spaces
        if entropy is None:
            entropy = -log_prob.mean()

        return log_prob, entropy, values

    def get_action(
        self,
        obs: ArrayOrTensor,
        action_mask: Optional[ArrayOrTensor] = None,
        hidden_state: Optional[dict[str, ArrayOrTensor]] = None,
    ) -> Union[
        tuple[
            np.ndarray,  # action
            np.ndarray,  # log_prob
            np.ndarray,  # entropy
            np.ndarray,  # values
            Optional[dict[str, ArrayOrTensor]],  # next_hidden_state
        ],
        tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],  # non-recurrent case
    ]:
        """Returns the next action to take in the environment.

        :param obs: Environment observation, or multiple observations in a batch
        :type obs: ArrayOrTensor
        :param action_mask: Mask of legal actions 1=legal 0=illegal, defaults to None
        :type action_mask: Optional[ArrayOrTensor]
        :param hidden_state: Hidden state for recurrent policies, defaults to None
        :type hidden_state: Optional[dict[str, ArrayOrTensor]]
        :return: Action, log probability, entropy, state values, and (if recurrent) next hidden state
        :rtype: Union[tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, Optional[dict[str, ArrayOrTensor]]], tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]]
        """
        obs = self.preprocess_observation(obs)
        with torch.no_grad():
            (
                action,
                log_prob,
                entropy,
                values,
                next_hidden,
            ) = self._get_action_and_values(
                obs,
                action_mask,
                hidden_state,
                sample=True,  # Explicitly sample=True during get_action
            )

        # Use -log_prob as entropy when squashing output in continuous action spaces
        entropy = -log_prob.mean() if entropy is None else entropy

        # Clip to action space during inference
        action_np = action.cpu().data.numpy()
        if not self.training and isinstance(self.action_space, spaces.Box):
            if self.actor.squash_output:
                action_np = self.actor.scale_action(action_np)
            else:
                action_np = np.clip(
                    action_np, self.action_space.low, self.action_space.high
                )

        log_prob_np = log_prob.cpu().data.numpy()
        entropy_np = entropy.cpu().data.numpy()
        values_np = values.cpu().data.numpy()

        if self.recurrent:
            return (
                action_np,
                log_prob_np,
                entropy_np,
                values_np,
                next_hidden if next_hidden is not None else None,
            )
        else:
            return (
                action_np,
                log_prob_np,
                entropy_np,
                values_np,
            )

    def learn(self, experiences: Optional[ExperiencesType] = None) -> float:
        """Updates agent network parameters to learn from experiences.

        :param experiences: Tuple of batched states, actions, log_probs, rewards, dones, values, next_state, next_done.
                            If use_rollout_buffer=True and experiences=None, uses data from rollout buffer.
        :type experiences: Optional[ExperiencesType]
        :return: Mean loss value from training.
        :rtype: float
        """
        if self.use_rollout_buffer:
            # NOTE: we are still allowing experiences to be passed in for backwards compatibility
            # but we will remove this in a future releases.
            # i.e. it's possible to do one learn with rollouts, then another with experiences on the same agent
            if experiences is None:
                # Learn from the internal rollout buffer
                if self.recurrent:
                    return self._learn_from_rollout_buffer_bptt()
                else:
                    return self._learn_from_rollout_buffer_flat()

        return self._deprecated_learn_from_experiences(experiences)

    def _deprecated_learn_from_experiences(self, experiences: ExperiencesType) -> float:
        """Deprecated method for learning from experiences tuple format.

        This method is deprecated and will be removed in a future release. The PPO implementation
        now uses a rollout buffer for improved performance, cleaner support for recurrent policies,
        and easier integration with custom environments.

        To migrate:
        1. Set use_rollout_buffer=True when creating PPO agent
        2. If using recurrent policies, set recurrent=True
        3. Use collect_rollouts() to gather experiences instead of passing experiences tuple
        4. Call learn() without arguments to train on collected rollouts
        """
        if not experiences:
            raise ValueError(
                "Experiences must be provided when use_rollout_buffer is False"
            )

        # Not self.use_rollout_buffer
        (
            observations,
            actions,
            log_probs,
            rewards,
            dones,
            values,
            next_obs,
            next_done,
        ) = stack_experiences(*experiences)

        # Bootstrapping returns using GAE advantage estimation
        dones = dones.long()
        with torch.no_grad():
            num_steps = rewards.size(0)
            next_obs = self.preprocess_observation(next_obs)
            next_value = self.critic(next_obs).reshape(1, -1).cpu()
            advantages = torch.zeros_like(rewards).float()
            last_gae_lambda = 0
            for t in reversed(range(num_steps)):
                if t == num_steps - 1:
                    next_non_terminal = 1.0 - next_done
                    nextvalue = next_value.squeeze()
                else:
                    next_non_terminal = 1.0 - dones[t + 1]
                    nextvalue = values[t + 1]

                # Calculate delta (TD error)
                delta = (
                    rewards[t] + self.gamma * nextvalue * next_non_terminal - values[t]
                )

                # Use recurrence relation to compute advantage
                advantages[t] = last_gae_lambda = (
                    delta
                    + self.gamma * self.gae_lambda * next_non_terminal * last_gae_lambda
                )

            returns = advantages + values

        # Flatten experiences from (batch_size, num_envs, ...) to (batch_size*num_envs, ...)
        # after checking if experiences are vectorized
        experiences = (observations, actions, log_probs, advantages, returns, values)
        if is_vectorized_experiences(*experiences):
            experiences = flatten_experiences(*experiences)

        # Move experiences to algo device
        experiences = self.to_device(*experiences)

        # Get number of samples from the returns tensor
        num_samples = experiences[4].size(0)
        batch_idxs = np.arange(num_samples)
        mean_loss = 0
        for _ in range(self.update_epochs):
            np.random.shuffle(batch_idxs)
            for start in range(0, num_samples, self.batch_size):
                minibatch_idxs = batch_idxs[start : start + self.batch_size]
                (
                    batch_observations,
                    batch_actions,
                    batch_log_probs,
                    batch_advantages,
                    batch_returns,
                    batch_values,
                ) = get_experiences_samples(minibatch_idxs, *experiences)

                batch_actions = batch_actions.squeeze()
                batch_returns = batch_returns.squeeze()
                batch_log_probs = batch_log_probs.squeeze()
                batch_advantages = batch_advantages.squeeze()
                batch_values = batch_values.squeeze()

                if len(minibatch_idxs) > 1:
                    log_prob, entropy, value = self.evaluate_actions(
                        obs=batch_observations, actions=batch_actions, hidden_state=None
                    )

                    logratio = log_prob - batch_log_probs
                    ratio = logratio.exp()

                    with torch.no_grad():
                        approx_kl = ((ratio - 1) - logratio).mean()

                    minibatch_advs = batch_advantages
                    minibatch_advs = (minibatch_advs - minibatch_advs.mean()) / (
                        minibatch_advs.std() + 1e-8
                    )

                    # Policy loss
                    pg_loss1 = -minibatch_advs * ratio
                    pg_loss2 = -minibatch_advs * torch.clamp(
                        ratio, 1 - self.clip_coef, 1 + self.clip_coef
                    )

                    pg_loss = torch.max(pg_loss1, pg_loss2).mean()

                    # Value loss
                    value = value.view(-1)
                    v_loss_unclipped = (value - batch_returns) ** 2
                    v_clipped = batch_values + torch.clamp(
                        value - batch_values, -self.clip_coef, self.clip_coef
                    )

                    v_loss_clipped = (v_clipped - batch_returns) ** 2
                    v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)
                    v_loss = 0.5 * v_loss_max.mean()

                    entropy_loss = entropy.mean()
                    loss = (
                        pg_loss - self.ent_coef * entropy_loss + v_loss * self.vf_coef
                    )

                    # actor + critic loss backprop
                    self.optimizer.zero_grad()
                    if self.accelerator is not None:
                        self.accelerator.backward(loss)
                    else:
                        loss.backward()

                    # Clip gradients
                    clip_grad_norm_(self.actor.parameters(), self.max_grad_norm)
                    clip_grad_norm_(self.critic.parameters(), self.max_grad_norm)

                    self.optimizer.step()

                    mean_loss += loss.item()

            if self.target_kl is not None:
                if approx_kl > self.target_kl:
                    break

        mean_loss /= num_samples * self.update_epochs
        return mean_loss

    def _learn_from_rollout_buffer_flat(
        self, buffer_td_external: Optional[TensorDict] = None
    ) -> float:
        """Learning procedure using flattened samples (no BPTT)."""
        if buffer_td_external is not None:
            buffer_td = buffer_td_external
        else:
            # .get_tensor_batch() returns a TensorDict on the specified device
            buffer_td = self.rollout_buffer.get_tensor_batch(device=self.device)

        if buffer_td.is_empty():
            warnings.warn("Buffer data is empty. Skipping learning step.")
            return 0.0

        # Normalize advantages globally
        valid_advantages: torch.Tensor = buffer_td.get("advantages")
        normalized_advantages = (valid_advantages - valid_advantages.mean()) / (
            valid_advantages.std() + 1e-8
        )
        buffer_td["advantages"] = normalized_advantages

        batch_size = self.batch_size
        num_samples = self.rollout_buffer.size()
        indices = np.arange(num_samples)
        mean_loss = 0.0
        approx_kl_divs = []
        for _ in range(self.update_epochs):
            np.random.shuffle(indices)
            for start_idx in range(0, num_samples, batch_size):
                end_idx = min(start_idx + batch_size, num_samples)
                minibatch_indices = indices[start_idx:end_idx]

                # Slice the TensorDict to get the minibatch
                minibatch_td = buffer_td[minibatch_indices]

                mb_obs = minibatch_td["observations"]
                mb_actions = minibatch_td["actions"]
                mb_log_probs = minibatch_td["log_probs"]
                mb_advantages = minibatch_td["advantages"]
                mb_returns = minibatch_td["returns"]
                mb_old_values = minibatch_td["values"]

                if isinstance(self.action_space, spaces.Discrete):
                    mb_actions = mb_actions.squeeze(-1)

                log_probs, entropy, values = self.evaluate_actions(
                    obs=mb_obs, actions=mb_actions, hidden_state=None
                )

                # Policy loss
                ratio = torch.exp(log_probs - mb_log_probs)
                policy_loss1 = -mb_advantages * ratio
                policy_loss2 = -mb_advantages * torch.clamp(
                    ratio, 1 - self.clip_coef, 1 + self.clip_coef
                )
                policy_loss = torch.max(policy_loss1, policy_loss2).mean()

                # Value loss
                value = values.view(-1)
                v_loss_unclipped = (value - mb_returns) ** 2
                v_clipped = mb_old_values + torch.clamp(
                    value - mb_old_values, -self.clip_coef, self.clip_coef
                )

                v_loss_clipped = (v_clipped - mb_returns) ** 2
                v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)
                v_loss = 0.5 * v_loss_max.mean()

                # Entropy loss
                entropy_loss = -entropy.mean()

                # Total loss
                loss = (
                    policy_loss + self.vf_coef * v_loss + self.ent_coef * entropy_loss
                )

                with torch.no_grad():
                    log_ratio = log_probs - mb_log_probs
                    approx_kl = ((ratio - 1) - log_ratio).mean().item()
                    approx_kl_divs.append(approx_kl)

                self.optimizer.zero_grad()
                if self.accelerator is not None:
                    self.accelerator.backward(loss)
                else:
                    loss.backward()

                clip_grad_norm_(self.actor.parameters(), self.max_grad_norm)
                clip_grad_norm_(self.critic.parameters(), self.max_grad_norm)

                self.optimizer.step()

                mean_loss += loss.item()

            if self.target_kl is not None and np.mean(approx_kl_divs) > self.target_kl:
                break  # Early stopping for the epoch if KL divergence target is exceeded

        mean_loss /= num_samples * self.update_epochs
        return mean_loss

    def _learn_from_rollout_buffer_bptt(self) -> float:
        """Learning procedure using truncated BPTT for recurrent networks.

        :return: Mean loss over the epochs
        :rtype: float
        """
        buffer_size = (
            self.rollout_buffer.capacity
            if self.rollout_buffer.full
            else self.rollout_buffer.pos
        )

        # Normalize advantages globally
        valid_advantages: torch.Tensor = self.rollout_buffer.buffer.get("advantages")[
            :buffer_size
        ]
        original_shape = valid_advantages.shape
        flat_adv = valid_advantages.reshape(-1)
        normalized_flat_adv = (flat_adv - flat_adv.mean()) / (flat_adv.std() + 1e-8)
        self.rollout_buffer.buffer["advantages"][:buffer_size] = (
            normalized_flat_adv.reshape(original_shape)
        )

        # Form padded sequences to perform BPTT on
        self.rollout_buffer.prepare_sequence_tensors(device=self.device)

        # Here, batch_size means number of sequences per minibatch
        mean_loss = 0.0
        total_minibatch_updates_total = 0
        for epoch in range(self.update_epochs):
            approx_kl_divs_epoch = []  # KL divergences for this epoch's minibatches
            num_minibatches_this_epoch = 0

            # Itreate over minibatches of sequences
            minibatch_gen = self.rollout_buffer.get_minibatch_sequences(
                batch_size=self.batch_size,
            )
            for minibatch_padded, minibatch_unpadded in minibatch_gen:
                # Obs shape: (batch_seq * seq_len, *obs_dims) or nested TD
                # Actions shape: (batch_seq * seq_len, *act_dims)
                # Other tensors shape: (batch_seq * seq_len, )
                mb_obs_seq = minibatch_padded["observations"]
                mb_actions_seq = minibatch_padded["actions"]
                mb_pad_mask = minibatch_padded["pad_mask"]
                mb_old_log_probs = minibatch_unpadded["log_probs"]
                mb_advantages = minibatch_unpadded["advantages"]
                mb_values = minibatch_unpadded["values"]
                mb_returns = minibatch_unpadded["returns"]
                mb_initial_hidden_states_dict: dict[str, torch.Tensor] = (
                    minibatch_padded.get_non_tensor(
                        "initial_hidden_states", default=None
                    )
                )

                approx_kl_divs_minibatch_timesteps = []

                # For actor: {key: (layers, batch_seq_size, hidden_size)}
                if self.recurrent and mb_initial_hidden_states_dict is not None:
                    mb_initial_hidden_states_dict = {
                        # val is (batch_seq_size, layers, size), permute to (layers, batch_seq_size, size)
                        key: val.permute(1, 0, 2).contiguous().to(self.device)
                        for key, val in mb_initial_hidden_states_dict.items()
                    }

                # Need to flatten action dimension for Discrete action spaces
                if isinstance(self.action_space, spaces.Discrete):
                    mb_actions_seq = mb_actions_seq.squeeze(-1)

                # new_value: (batch_seq,),
                # entropy: (batch_seq,) or scalar,
                # log_prob: (batch_seq,)
                (
                    new_log_probs,
                    entropy,
                    new_values,
                ) = self.evaluate_actions(
                    obs=mb_obs_seq,
                    actions=mb_actions_seq,
                    hidden_state=mb_initial_hidden_states_dict,
                )

                # Mask out padded values
                new_values = new_values[mb_pad_mask]
                new_log_probs = new_log_probs[mb_pad_mask]
                entropy = entropy[mb_pad_mask]

                if isinstance(entropy, torch.Tensor):
                    entropy = entropy.mean()

                # Policy loss
                ratio = torch.exp(new_log_probs - mb_old_log_probs)
                policy_loss1 = -mb_advantages * ratio
                policy_loss2 = -mb_advantages * torch.clamp(
                    ratio, 1 - self.clip_coef, 1 + self.clip_coef
                )
                policy_loss = torch.max(policy_loss1, policy_loss2).mean()

                # Value loss
                v_loss_unclipped = (new_values - mb_returns) ** 2
                v_clipped = mb_values + torch.clamp(
                    new_values - mb_values, -self.clip_coef, self.clip_coef
                )

                v_loss_clipped = (v_clipped - mb_returns) ** 2
                v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)
                value_loss = v_loss_max.mean()

                # Entropy loss
                entropy_loss = -entropy

                with torch.no_grad():
                    log_ratio = new_log_probs - mb_old_log_probs
                    approx_kl_divs_minibatch_timesteps.append(
                        ((torch.exp(log_ratio) - 1) - log_ratio).mean().item()
                    )

                loss = (
                    policy_loss
                    + self.vf_coef * value_loss
                    + self.ent_coef * entropy_loss
                )

                self.optimizer.zero_grad()
                if self.accelerator is not None:
                    self.accelerator.backward(loss)
                else:
                    loss.backward()

                clip_grad_norm_(self.actor.parameters(), self.max_grad_norm)
                clip_grad_norm_(self.critic.parameters(), self.max_grad_norm)
                self.optimizer.step()

                mean_loss += loss.item()
                num_minibatches_this_epoch += 1

                if (
                    self.target_kl is not None
                    and len(approx_kl_divs_minibatch_timesteps) > 0
                ):
                    # Average KL over all timesteps in this minibatch of sequences
                    kl_for_current_minibatch = np.mean(
                        approx_kl_divs_minibatch_timesteps
                    )
                    approx_kl_divs_epoch.append(
                        kl_for_current_minibatch
                    )  # Store minibatch average KL

                    if kl_for_current_minibatch > self.target_kl:
                        warnings.warn(
                            f"Epoch {epoch}, Minibatch: KL divergence {kl_for_current_minibatch:.4f} exceeded target {self.target_kl}. Stopping update for this epoch."
                        )
                        break  # Break from minibatch loop for this epoch

            total_minibatch_updates_total += num_minibatches_this_epoch
            # Check average KL for the epoch if target_kl is set and the inner loop wasn't broken by KL
            if self.target_kl is not None and len(approx_kl_divs_epoch) > 0:
                avg_kl_this_epoch = np.mean(approx_kl_divs_epoch)
                if (
                    avg_kl_this_epoch > self.target_kl
                    and not (  # Ensure this wasn't the break from inner loop
                        len(approx_kl_divs_minibatch_timesteps) > 0
                        and np.mean(approx_kl_divs_minibatch_timesteps) > self.target_kl
                    )
                ):
                    warnings.warn(
                        f"Epoch {epoch}: Average KL divergence {avg_kl_this_epoch:.4f} exceeded target {self.target_kl} after completing epoch. Consider adjusting learning rate or target_kl."
                    )
                    # This break is for the epoch loop if KL was exceeded on average for the epoch
                    # but not necessarily in the last minibatch that would have broken the inner loop.
                    break

            # If inner loop broke due to KL, this outer break also executes
            if (
                self.target_kl is not None
                and len(approx_kl_divs_minibatch_timesteps) > 0
                and np.mean(approx_kl_divs_minibatch_timesteps) > self.target_kl
            ):
                break

        mean_loss = mean_loss / max(1e-8, total_minibatch_updates_total)
        return mean_loss

    def test(
        self,
        env: GymEnvType,
        swap_channels: bool = False,
        max_steps: Optional[int] = None,
        loop: int = 3,
        vectorized: bool = True,
        callback: Optional[Callable[[float, dict[str, float]], None]] = None,
    ) -> float:
        """Returns mean test score of agent in environment with epsilon-greedy policy.

        :param env: The environment to be tested in
        :type env: GymEnvType
        :param swap_channels: Swap image channels dimension from last to first [H, W, C] -> [C, H, W], defaults to False
        :type swap_channels: bool, optional
        :param max_steps: Maximum number of testing steps, defaults to None
        :type max_steps: int, optional
        :param loop: Number of testing loops/episodes to complete. The returned score is the mean. Defaults to 3
        :type loop: int, optional
        :param vectorized: Whether the environment is vectorized, defaults to True
        :type vectorized: bool, optional
        :param callback: Optional callback function that takes the sum of rewards and the last info dictionary as input, defaults to None
        :type callback: Optional[Callable[[float, dict[str, float]], None]]

        :return: Mean test score of agent in environment
        :rtype: float
        """
        # set to evaluation mode. This is important for batch norm and dropout layers
        self.actor.eval()
        self.critic.eval()
        self.set_training_mode(False)

        with torch.no_grad():
            rewards = []
            num_envs = env.num_envs if hasattr(env, "num_envs") and vectorized else 1

            for _ in range(loop):
                obs, info = env.reset()
                scores = np.zeros(num_envs)
                completed_episode_scores = np.zeros(num_envs)
                finished = np.zeros(num_envs, dtype=bool)
                step = 0
                test_hidden_state = (
                    self.get_initial_hidden_state(num_envs) if self.recurrent else None
                )

                last_infos = (
                    [{}] * num_envs if vectorized else {}
                )  # Initialize last_info holder

                while not np.all(finished):
                    if swap_channels:
                        obs = obs_channels_to_first(obs)

                    # Process action mask
                    action_mask = None
                    if vectorized:
                        # Check if info is a list/array of dicts
                        if (
                            isinstance(info, (list, np.ndarray))
                            and len(info) == num_envs
                            and all(isinstance(i, dict) for i in info)
                        ):
                            masks = [env_info.get("action_mask") for env_info in info]
                            # If all environments returned a mask and they are not None
                            if all(m is not None for m in masks):
                                try:
                                    action_mask = np.stack(masks)
                                except Exception as e:
                                    warnings.warn(f"Could not stack action masks: {e}")
                                    action_mask = None
                            # If only some environments returned masks, we probably can't use them reliably
                            elif any(m is not None for m in masks):
                                warnings.warn(
                                    "Action masks not provided for all vectorized environments. Skipping mask."
                                )
                                action_mask = None
                        # Handle case where info might be a single dict even if vectorized (e.g. VecNormalize)
                        elif isinstance(info, dict):
                            action_mask = info.get("action_mask", None)

                    else:  # Not vectorized
                        if isinstance(info, dict):
                            action_mask = info.get("action_mask", None)

                    # Get action
                    if self.recurrent:
                        action, _, _, _, test_hidden_state = self.get_action(
                            obs, action_mask=action_mask, hidden_state=test_hidden_state
                        )
                    else:
                        action, _, _, _ = self.get_action(obs, action_mask=action_mask)

                    # Environment step
                    if vectorized:
                        obs, reward, done, trunc, info = env.step(action)
                        last_infos = info  # Store the array of infos
                    else:
                        obs, reward, done, trunc, info_single = env.step(action[0])
                        # Store info in a dictionary for consistency if not vectorized
                        info = {"final_info": info_single} if done or trunc else {}
                        last_infos = info  # Store the single info dict

                    step += 1
                    scores += np.array(reward)

                    # Check for episode termination
                    newly_finished = (
                        np.logical_or(
                            np.logical_or(done, trunc),
                            (max_steps is not None and step == max_steps),
                        )
                        & ~finished
                    )

                    # Reset hidden state for newly finished environments
                    if self.recurrent and np.any(newly_finished):
                        initial_hidden_states_for_reset = self.get_initial_hidden_state(
                            num_envs
                        )
                        if isinstance(test_hidden_state, dict):
                            for key in test_hidden_state:
                                reset_states = initial_hidden_states_for_reset[key][
                                    :, newly_finished, :
                                ]
                                if reset_states.shape[1] > 0:
                                    test_hidden_state[key][
                                        :, newly_finished, :
                                    ] = reset_states

                    if np.any(newly_finished):
                        completed_episode_scores[newly_finished] = scores[
                            newly_finished
                        ]
                        finished[newly_finished] = True

                # End of episode loop for one test run
                loop_reward_sum = np.sum(completed_episode_scores)

                # Prepare info for callback
                final_info_for_callback = {}
                if vectorized:
                    if (
                        isinstance(last_infos, (list, np.ndarray))
                        and len(last_infos) > 0
                    ):
                        final_info_for_callback = (
                            last_infos[0] if isinstance(last_infos[0], dict) else {}
                        )
                    elif isinstance(last_infos, dict):
                        final_info_for_callback = last_infos
                else:  # Not vectorized
                    if isinstance(last_infos, dict):
                        final_info_for_callback = last_infos

                if callback is not None:
                    callback(loop_reward_sum, final_info_for_callback)

                rewards.append(np.mean(completed_episode_scores))

        mean_fit = np.mean(rewards)
        self.fitness.append(mean_fit)

        # cleanup evaluation mode back into the default training mode (e.g. batch norm and dropout layers)
        self.set_training_mode(True)
        self.actor.train()
        self.critic.train()

        return mean_fit



import time
import warnings
from datetime import datetime
from typing import Any, Optional, Union

import gymnasium as gym
import numpy as np
import wandb
from accelerate import Accelerator
from tensordict import TensorDictBase
from torch.utils.data import DataLoader

from agilerl.algorithms import DDPG, DQN, TD3, RainbowDQN
from agilerl.algorithms.core.base import RLAlgorithm
from agilerl.components import (
    MultiStepReplayBuffer,
    PrioritizedReplayBuffer,
    ReplayBuffer,
)
from agilerl.components.data import ReplayDataset, Transition
from agilerl.components.sampler import Sampler
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.utils.algo_utils import obs_channels_to_first
from agilerl.utils.utils import (
    default_progress_bar,
    init_wandb,
    save_population_checkpoint,
    tournament_selection_and_mutation,
)

InitDictType = Optional[dict[str, Any]]
PopulationType = list[RLAlgorithm]
BufferType = Union[ReplayBuffer, PrioritizedReplayBuffer, MultiStepReplayBuffer]


def train_off_policy(
    env: gym.Env,
    env_name: str,
    algo: str,
    pop: PopulationType,
    memory: BufferType,
    INIT_HP: InitDictType = None,
    MUT_P: InitDictType = None,
    swap_channels: bool = False,
    max_steps: int = 1000000,
    evo_steps: int = 10000,
    eval_steps: Optional[int] = None,
    eval_loop: int = 1,
    learning_delay: int = 0,
    eps_start: float = 1.0,
    eps_end: float = 0.1,
    eps_decay: float = 0.995,
    target: Optional[float] = None,
    n_step: bool = False,
    per: bool = False,
    n_step_memory: Optional[MultiStepReplayBuffer] = None,
    tournament: Optional[TournamentSelection] = None,
    mutation: Optional[Mutations] = None,
    checkpoint: Optional[int] = None,
    checkpoint_path: Optional[str] = None,
    overwrite_checkpoints: bool = False,
    save_elite: bool = False,
    elite_path: Optional[str] = None,
    wb: bool = False,
    verbose: bool = True,
    accelerator: Optional[Accelerator] = None,
    wandb_api_key: Optional[str] = None,
    wandb_kwargs: Optional[dict[str, Any]] = None,
) -> tuple[PopulationType, list[list[float]]]:
    """The general online RL training function. Returns trained population of agents
    and their fitnesses.

    :param env: The environment to train in. Can be vectorized.
    :type env: Gym-style environment
    :param env_name: Environment name
    :type env_name: str
    :param algo: RL algorithm name
    :type algo: str
    :param pop: Population of agents
    :type pop: list[object]
    :param memory: Experience Replay Buffer
    :type memory: object
    :param INIT_HP: Dictionary containing initial hyperparameters, defaults to None
    :type INIT_HP: dict, optional
    :param MUT_P: Dictionary containing mutation parameters, defaults to None
    :type MUT_P: dict, optional
    :param swap_channels: Swap image channels dimension from last to first
        [H, W, C] -> [C, H, W], defaults to False
    :type swap_channels: bool, optional
    :param max_steps: Maximum number of steps in environment, defaults to 1000000
    :type max_steps: int, optional
    :param evo_steps: Evolution frequency (steps), defaults to 10000
    :type evo_steps: int, optional
    :param eval_steps: Number of evaluation steps per episode. If None, will evaluate until
        environment terminates or truncates. Defaults to None
    :type eval_steps: int, optional
    :param eval_loop: Number of evaluation episodes, defaults to 1
    :type eval_loop: int, optional
    :param learning_delay: Steps in environment before starting learning, defaults to 0
    :type learning_delay: int, optional
    :param eps_start: Maximum exploration - initial epsilon value, defaults to 1.0
    :type eps_start: float, optional
    :param eps_end: Minimum exploration - final epsilon value, defaults to 0.1
    :type eps_end: float, optional
    :param eps_decay: Epsilon decay per episode, defaults to 0.995
    :type eps_decay: float, optional
    :param target: Target score for early stopping, defaults to None
    :type target: float, optional
    :param n_step: Use multi-step experience replay buffer, defaults to False
    :type n_step: bool, optional
    :param per: Using prioritized experience replay buffer, defaults to False
    :type per: bool, optional
    :param memory: Multi-step Experience Replay Buffer to be used alongside Prioritized
        ERB, defaults to None
    :type memory: object, optional
    :param tournament: Tournament selection object, defaults to None
    :type tournament: object, optional
    :param mutation: Mutation object, defaults to None
    :type mutation: object, optional
    :param checkpoint: Checkpoint frequency (steps), defaults to None
    :type checkpoint: int, optional
    :param checkpoint_path: Location to save checkpoint, defaults to None
    :type checkpoint_path: str, optional
    :param overwrite_checkpoints: Overwrite previous checkpoints during training, defaults to False
    :type overwrite_checkpoints: bool, optional
    :param save_elite: Boolean flag indicating whether to save elite member at the end
        of training, defaults to False
    :type save_elite: bool, optional
    :param elite_path: Location to save elite agent, defaults to None
    :type elite_path: str, optional
    :param wb: Weights & Biases tracking, defaults to False
    :type wb: bool, optional
    :param verbose: Display training stats, defaults to True
    :type verbose: bool, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wandb_api_key: API key for Weights & Biases, defaults to None
    :param wandb_kwargs: Additional kwargs to pass to wandb.init()
    :type wand_kwargs: dict, optional
    """
    assert isinstance(
        algo, str
    ), "'algo' must be the name of the algorithm as a string."
    assert isinstance(max_steps, int), "Number of steps must be an integer."
    assert isinstance(evo_steps, int), "Evolution frequency must be an integer."
    assert isinstance(eps_start, float), "Starting epsilon must be a float."
    assert isinstance(eps_end, float), "Final value of epsilon must be a float."
    assert isinstance(eps_decay, float), "Epsilon decay rate must be a float."
    if target is not None:
        assert isinstance(
            target, (float, int)
        ), "Target score must be a float or an integer."
    assert isinstance(n_step, bool), "'n_step' must be a boolean."
    assert isinstance(per, bool), "'per' must be a boolean."
    if checkpoint is not None:
        assert isinstance(checkpoint, int), "Checkpoint must be an integer."
    assert isinstance(
        wb, bool
    ), "'wb' must be a boolean flag, indicating whether to record run with W&B"
    assert isinstance(verbose, bool), "Verbose must be a boolean."
    if save_elite is False and elite_path is not None:
        warnings.warn(
            "'save_elite' set to False but 'elite_path' has been defined, elite will not\
                      be saved unless 'save_elite' is set to True."
        )
    if checkpoint is None and checkpoint_path is not None:
        warnings.warn(
            "'checkpoint' set to None but 'checkpoint_path' has been defined, checkpoint will not\
                      be saved unless 'checkpoint' is defined."
        )

    if wb:
        init_wandb_kwargs = {
            "algo": algo,
            "env_name": env_name,
            "init_hyperparams": INIT_HP,
            "mutation_hyperparams": MUT_P,
            "wandb_api_key": wandb_api_key,
            "accelerator": accelerator,
        }
        if wandb_kwargs is not None:
            init_wandb_kwargs.update(wandb_kwargs)
        init_wandb(**init_wandb_kwargs)

    # Detect if environment is vectorised
    if hasattr(env, "num_envs"):
        num_envs = env.num_envs
        is_vectorised = True
    else:
        num_envs = 1
        is_vectorised = False

    save_path = (
        checkpoint_path.split(".pt")[0]
        if checkpoint_path is not None
        else "{}-EvoHPO-{}-{}".format(
            env_name, algo, datetime.now().strftime("%m%d%Y%H%M%S")
        )
    )

    if accelerator is not None:
        # Create dataloader from replay buffer
        replay_dataset = ReplayDataset(memory, pop[0].batch_size)
        replay_dataloader = DataLoader(replay_dataset, batch_size=None)
        replay_dataloader = accelerator.prepare(replay_dataloader)
        sampler = Sampler(dataset=replay_dataset, dataloader=replay_dataloader)
    else:
        sampler = Sampler(memory=memory)
        if n_step_memory is not None:
            n_step_sampler = Sampler(memory=n_step_memory)

    if accelerator is not None:
        print(f"\nDistributed training on {accelerator.device}...")
    else:
        print("\nTraining...")

    # Format progress bar
    pbar = default_progress_bar(max_steps, accelerator)

    pop_loss = [[] for _ in pop]
    pop_fitnesses = []
    total_steps = 0
    loss = None
    checkpoint_count = 0

    # Pre-training mutation
    if accelerator is None and mutation is not None:
        pop = mutation.mutation(pop, pre_training_mut=True)

    # RL training loop
    while np.less([agent.steps[-1] for agent in pop], max_steps).all():
        if accelerator is not None:
            accelerator.wait_for_everyone()

        pop_episode_scores = []
        pop_fps = []
        for agent_idx, agent in enumerate(pop):  # Loop through population
            state, info = env.reset()  # Reset environment at start of episode
            scores = np.zeros(num_envs)
            completed_episode_scores, losses = [], []
            steps = 0

            if isinstance(agent, (DQN, RainbowDQN)):
                train_actions_hist = [0] * agent.action_dim

            if isinstance(agent, DQN):
                epsilon = eps_start

            start_time = time.time()
            for idx_step in range(evo_steps // num_envs):
                if swap_channels:
                    state = obs_channels_to_first(state)

                # Get next action from agent
                if isinstance(agent, DQN):
                    action_mask = info.get("action_mask", None)
                    action = agent.get_action(state, epsilon, action_mask=action_mask)
                    # Decay epsilon for exploration
                    epsilon = max(eps_end, epsilon * eps_decay)
                elif isinstance(agent, RainbowDQN):
                    action_mask = info.get("action_mask", None)
                    action = agent.get_action(state, action_mask=action_mask)
                else:
                    action = agent.get_action(state)

                if isinstance(agent, (DQN, RainbowDQN)):
                    for a in action:
                        if not isinstance(a, int):
                            a = int(a)
                        train_actions_hist[a] += 1

                if not is_vectorised:
                    action = action[0]

                # Act in environment
                next_state, reward, done, trunc, info = env.step(action)
                scores += np.array(reward)

                if not is_vectorised:
                    done = np.array([done])
                    trunc = np.array([trunc])

                reset_noise_indices = []
                for idx, (d, t) in enumerate(zip(done, trunc)):
                    if d or t:
                        completed_episode_scores.append(scores[idx])
                        agent.scores.append(scores[idx])
                        scores[idx] = 0
                        reset_noise_indices.append(idx)

                if isinstance(agent, (DDPG, TD3)):
                    agent.reset_action_noise(reset_noise_indices)

                total_steps += num_envs
                steps += num_envs

                # Save experience to replay buffer
                next_state = (
                    obs_channels_to_first(next_state) if swap_channels else next_state
                )

                transition: TensorDictBase = Transition(
                    obs=state,
                    action=action,
                    reward=reward,
                    next_obs=next_state,
                    done=done,
                )
                if not is_vectorised:
                    transition = transition.unsqueeze(0)

                # Add transition to replay buffer
                transition = transition.to_tensordict()
                transition.batch_size = [num_envs]
                if n_step_memory is not None:
                    one_step_transition = n_step_memory.add(transition)
                    if one_step_transition is not None:
                        memory.add(one_step_transition)
                else:
                    memory.add(transition)
                if per:
                    fraction = min(
                        ((agent.steps[-1] + idx_step + 1) * num_envs / max_steps), 1.0
                    )
                    agent.beta += fraction * (1.0 - agent.beta)

                # Learn according to learning frequency
                # Handle learn_step > num_envs
                if agent.learn_step > num_envs:
                    learn_step = agent.learn_step // num_envs
                    if (
                        idx_step % learn_step == 0
                        and len(memory) >= agent.batch_size
                        and memory.size > learning_delay
                    ):
                        if per:
                            experiences = sampler.sample(agent.batch_size, agent.beta)
                            if n_step_memory is not None:
                                n_step_experiences = n_step_sampler.sample(
                                    experiences["idxs"]
                                )
                            else:
                                n_step_experiences = None

                            loss, idxs, priorities = agent.learn(
                                experiences, n_experiences=n_step_experiences, per=per
                            )
                            memory.update_priorities(idxs, priorities)
                        else:
                            experiences = sampler.sample(
                                agent.batch_size,
                                return_idx=True if n_step_memory is not None else False,
                            )
                            if n_step_memory is not None:
                                n_step_experiences = n_step_sampler.sample(
                                    experiences["idxs"]
                                )
                                loss, *_ = agent.learn(
                                    experiences, n_experiences=n_step_experiences
                                )
                            else:
                                loss = agent.learn(experiences)
                                if isinstance(agent, RainbowDQN):
                                    loss, *_ = loss

                elif len(memory) >= agent.batch_size and memory.size > learning_delay:
                    for _ in range(num_envs // agent.learn_step):
                        # Sample replay buffer
                        # Learn according to agent's RL algorithm
                        if per:
                            experiences = sampler.sample(agent.batch_size, agent.beta)
                            if n_step_memory is not None:
                                n_step_experiences = n_step_sampler.sample(
                                    experiences["idxs"]
                                )
                            else:
                                n_step_experiences = None
                            loss, idxs, priorities = agent.learn(
                                experiences, n_experiences=n_step_experiences, per=per
                            )
                            memory.update_priorities(idxs, priorities)
                        else:
                            experiences = sampler.sample(
                                agent.batch_size,
                                return_idx=True if n_step_memory is not None else False,
                            )
                            if n_step_memory is not None:
                                n_step_experiences = n_step_sampler.sample(
                                    experiences["idxs"]
                                )
                                loss, *_ = agent.learn(
                                    experiences, n_experiences=n_step_experiences
                                )
                            else:
                                loss = agent.learn(experiences)
                                if isinstance(agent, RainbowDQN):
                                    loss, *_ = loss

                if loss is not None:
                    losses.append(loss)

                state = next_state

            pbar.update(evo_steps // len(pop))

            agent.steps[-1] += steps
            fps = steps / (time.time() - start_time)
            pop_fps.append(fps)
            pop_episode_scores.append(completed_episode_scores)

            if len(losses) > 0:
                if isinstance(losses[-1], tuple):
                    actor_losses, critic_losses = list(zip(*losses))
                    mean_loss = np.mean(
                        [loss for loss in actor_losses if loss is not None]
                    ), np.mean(critic_losses)
                else:
                    mean_loss = np.mean(losses)

                pop_loss[agent_idx].append(mean_loss)

        if isinstance(agent, DQN):
            # Reset epsilon start to final epsilon value of this epoch
            eps_start = epsilon

        # Evaluate population
        fitnesses = [
            agent.test(
                env, swap_channels=swap_channels, max_steps=eval_steps, loop=eval_loop
            )
            for agent in pop
        ]
        pop_fitnesses.append(fitnesses)
        mean_scores = [
            (
                np.mean(episode_scores)
                if len(episode_scores) > 0
                else "0 completed episodes"
            )
            for episode_scores in pop_episode_scores
        ]

        if wb:
            wandb_dict = {
                "global_step": (
                    total_steps * accelerator.state.num_processes
                    if accelerator is not None and accelerator.is_main_process
                    else total_steps
                ),
                "fps": np.mean(pop_fps),
                "train/mean_score": np.mean(
                    [
                        mean_score
                        for mean_score in mean_scores
                        if not isinstance(mean_score, str)
                    ]
                ),
                "eval/mean_fitness": np.mean(fitnesses),
                "eval/best_fitness": np.max(fitnesses),
            }

            # Create the loss dictionaries
            if isinstance(agent, (DQN, RainbowDQN)):
                actor_loss_dict = {
                    f"train/agent_{index}_actor_loss": np.mean(loss[-10:])
                    for index, loss in enumerate(pop_loss)
                }
                wandb_dict.update(actor_loss_dict)
            elif isinstance(agent, (DDPG, TD3)):
                actor_loss_dict = {
                    f"train/agent_{index}_actor_loss": np.mean(
                        list(zip(*loss_list))[0][-10:]
                    )
                    for index, loss_list in enumerate(pop_loss)
                }
                critic_loss_dict = {
                    f"train/agent_{index}_critic_loss": np.mean(
                        list(zip(*loss_list))[-1][-10:]
                    )
                    for index, loss_list in enumerate(pop_loss)
                }
                wandb_dict.update(actor_loss_dict)
                wandb_dict.update(critic_loss_dict)

            if isinstance(agent, (DQN, RainbowDQN)):
                train_actions_hist = [
                    freq / sum(train_actions_hist) for freq in train_actions_hist
                ]
                train_actions_dict = {
                    f"train/action_{index}": action
                    for index, action in enumerate(train_actions_hist)
                }
                wandb_dict.update(train_actions_dict)

            if accelerator is not None:
                accelerator.wait_for_everyone()
                if accelerator.is_main_process:
                    wandb.log(wandb_dict)
                accelerator.wait_for_everyone()
            else:
                wandb.log(wandb_dict)

        # Update step counter
        for agent in pop:
            agent.steps.append(agent.steps[-1])

        # Early stop if consistently reaches target
        if target is not None:
            if (
                np.all(
                    np.greater([np.mean(agent.fitness[-10:]) for agent in pop], target)
                )
                and len(pop[0].steps) >= 100
            ):
                if wb:
                    wandb.finish()
                return pop, pop_fitnesses

        # Tournament selection and population mutation
        if tournament and mutation is not None:
            pop = tournament_selection_and_mutation(
                population=pop,
                tournament=tournament,
                mutation=mutation,
                env_name=env_name,
                algo=algo,
                elite_path=elite_path,
                save_elite=save_elite,
                accelerator=accelerator,
            )

        if verbose:
            fitness = ["%.2f" % fitness for fitness in fitnesses]
            avg_fitness = ["%.2f" % np.mean(agent.fitness[-5:]) for agent in pop]
            avg_score = ["%.2f" % np.mean(agent.scores[-10:]) for agent in pop]
            agents = [agent.index for agent in pop]
            num_steps = [agent.steps[-1] for agent in pop]
            muts = [agent.mut for agent in pop]

            banner_text = f"Global Steps {total_steps}"
            banner_width = max(len(banner_text) + 8, 35)
            border = "=" * banner_width
            centered_text = f"{banner_text}".center(banner_width)
            pbar.write(
                f"{border}\n"
                f"{centered_text}\n"
                f"{border}\n"
                f"Fitness:\t\t{fitness}\n"
                f"Score:\t\t{mean_scores}\n"
                f"5 fitness avgs:\t{avg_fitness}\n"
                f"10 score avgs:\t{avg_score}\n"
                f"Agents:\t\t{agents}\n"
                f"Steps:\t\t{num_steps}\n"
                f"Mutations:\t\t{muts}"
            )

        # Save model checkpoint
        if checkpoint is not None:
            if pop[0].steps[-1] // checkpoint > checkpoint_count:
                save_population_checkpoint(
                    population=pop,
                    save_path=save_path,
                    overwrite_checkpoints=overwrite_checkpoints,
                    accelerator=accelerator,
                )
                checkpoint_count += 1

    if wb:
        if accelerator is not None:
            accelerator.wait_for_everyone()
            if accelerator.is_main_process:
                wandb.finish()
            accelerator.wait_for_everyone()
        else:
            wandb.finish()

    pbar.close()
    return pop, pop_fitnesses
